package com.shellinabox.console;

import com.shellinabox.clipboard.Clipboard;
import com.shellinabox.clipboard.ClipboardOwner;
import com.shellinabox.emulation.VT102;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Scrollbar;
import java.awt.Toolkit;


public abstract class Console extends Canvas implements ClipboardOwner {

    // Disable double buffering if the performance hit is too big
    private static final boolean useDoubleBuffering = true;

    // Tweak these parameters to select how often the display is forced
    // to be updated
    private static final int CURSORPOS_WEIGHT       = 50;
    private static final int DRAWSTRING_WEIGHT      = 1;
    private static final int SCROLL_WEIGHT          = 900;
    private static final int CLEAR_WEIGHT           = 100;
    private static final int OPERATIONS_THRESHOLD   = 1000;
    private static final int FLUSH_THRESHOLD        = 5;

    private static final char CURSOR_SET_ATTR       = 0xF8; // Solid red block
    private static final char CURSOR_XOR_ATTR       = 0x37;
    private static final int[]vgaColors             = {0, 4, 2, 6, 1, 5, 3, 7,
                                                       8,12,10,14, 9,13,11,15};

    public static final int  DEFAULT_SCROLLBACK     = 500;
    public static final int  MAXIMUM_SCROLLBACK     = 2000;
    public static final char REVERSE                = '\u0100';
    public static final char UNDERLINE              = '\u0200';
    public static final char DIM                    = '\u0400';
    public static final char BRIGHT                 = '\u0800';
    public static final char BLINK                  = '\u1000';
    public static final char SELECTED               = '\u2000';

    private Object        repaintMonitor            = new Object();
    private Object        repaintDoneMonitor        = new Object();
    private Clipboard     clipboard;
    private VT102         emulator;
    private Image         consoleImage;
    private Graphics      consoleGraphics;
    private char[]        blanks;
    private Color[]       consoleColors;
    private boolean       isInvalid;
    private int           invalidX;
    private int           invalidY;
    private int           invalidW;
    private int           invalidH;
    private boolean       visualBell;
    private int           maxScrollBackLines;
    private Scrollbar     scrollbar;
    private int           operationsCounter;
    private int           flushCounter;
    private boolean       sizeIsValid;
    private char[][]      alternateCharBuffer;
    private char[][]      alternateAttrBuffer;
    private boolean       isInverted;
    private boolean       isAlternateScreen;
    private boolean       utfEnabled;
    private boolean       settingClipboard;
    private boolean       isSelected;
    private int           selectedX1;
    private int           selectedY1;
    private int           selectedX2;
    private int           selectedY2;
    
    ScrollMonitor         scrollMonitor;
    
    protected int         width                     = -1;
    protected int         height                    = -1;
    protected int         cellWidth;
    protected int         cellHeight;
    protected int         charWidth                 = -1;
    protected int         charHeight                = -1;
    protected boolean     cursorVisible             = true;
    protected int         cursorX;
    protected int         cursorY;
    protected char        cursorSetAttr             = CURSOR_SET_ATTR;
    protected char        cursorXORAttr             = CURSOR_XOR_ATTR;
    protected boolean     isFocused;
    protected char[][]    consoleCharBuffer;
    protected char[][]    consoleAttrBuffer;
    protected int         scrollBackLines;
    protected int         scrollOffset;


    abstract void repaintArea(Graphics graphics,int x,int y,int w,int h);
    abstract void repaintCursorInRectangle(Graphics graphics,
                                           int x,int y,int w,int h);
    abstract Dimension computeFontSize(Graphics graphics);


    public Console(int maxScrollBackLines,boolean visualBell) {
        this.maxScrollBackLines= maxScrollBackLines;
        this.visualBell        = visualBell;
        clipboard              = Clipboard.getSystemClipboard();
        consoleColors          = new Color[16];
        consoleColors[ 0]      = Color.black;
        consoleColors[ 1]      = Color.red;
        consoleColors[ 2]      = Color.green;
        consoleColors[ 3]      = Color.yellow;
        consoleColors[ 4]      = Color.blue;
        consoleColors[ 5]      = Color.magenta;
        consoleColors[ 6]      = Color.cyan;
        consoleColors[ 7]      = Color.lightGray;
        consoleColors[ 8]      = Color.darkGray;
        consoleColors[15]      = Color.white;
        for (int i = 1; i <= 6; i++) {
            float[] hsb        = Color.RGBtoHSB(consoleColors[i].getRed(),
                                                consoleColors[i].getGreen(),
                                                consoleColors[i].getBlue(),
                                                null);
            // add a little more black to the default colors to make them
            // more distinct from the light colors
            consoleColors[i]   = Color.getHSBColor(hsb[0],
                                                   hsb[1],
                                                   (float)(hsb[2] * 0.7));
            // remove some saturation in order to get light colors
            consoleColors[i+8] = Color.getHSBColor(hsb[0],
                                                   (float)(hsb[1] * 0.9),
                                                   hsb[2]);
        }
        scrollbar              = new ConsoleScrollbar(Scrollbar.VERTICAL);
        return;
    }


    public void registerEmulation(VT102 emulator) {
        this.emulator = emulator;
        if (sizeIsValid)
            emulator.setSize(new Dimension(charWidth,charHeight));
        return;
    }


    public Dimension getCharSize() {
        if (!sizeIsValid)
            return(null);
        return(new Dimension(charWidth,charHeight));
    }


    public Dimension getPixelSize() {
        if (!sizeIsValid)
            return(null);
        return(new Dimension(width,height));
    }


    public Scrollbar getScrollbar() {
        return(scrollbar);
    }


    public void paint(Graphics g) {
        synchronized (repaintMonitor) {
            repaintMonitor.notify();
        }
        Dimension curSize  = size();
        Rectangle clipRect = g.getClipRect();
        if (clipRect == null)
            clipRect = new Rectangle(0,0,curSize.width,curSize.height);
        synchronized (this) {
            // Reset console if size has changed since last repaint
            if (curSize.width  != width ||
                curSize.height != height)
                resetConsole(g,curSize.width,curSize.height);
            Graphics graphics;
            if (useDoubleBuffering)
                graphics = consoleGraphics;
            else
                graphics = g;
            graphics.setPaintMode();
            if (!useDoubleBuffering || isInvalid) {
                Rectangle fudge = hitClip(clipRect,0,charHeight*cellHeight,
                                          width,cellHeight);
                if (fudge != null) {
                    graphics.setColor(getBackground());
                    graphics.fillRect(fudge.x,fudge.y,
                                      fudge.width,fudge.height);
                }
                if (hitClip(clipRect,charWidth*cellWidth,0,cellWidth,
                            charHeight*cellHeight) != null) {
                    for (int y1 = scrollOffset; y1 < scrollOffset+charHeight;){
                        char attr = consoleAttrBuffer[y1][charWidth-1];
                        int y2 = y1;
                        while (++y2 < scrollOffset+charHeight &&
                               attr == consoleAttrBuffer[y2][charWidth-1]);
                        setAttrColors(graphics,attr);
                        graphics.fillRect(charWidth*cellWidth,
                                          (y1-scrollOffset)*cellHeight,
                                          cellWidth,(y2-y1)*cellHeight);
                        y1 = y2;
                    }
                }
                int x = Math.max(0,clipRect.x/cellWidth);
                int y = Math.max(0,clipRect.y/cellHeight);
                int w = Math.min(charWidth-x,
                                 (clipRect.x+clipRect.width +cellWidth -1)/
                                 cellWidth -x);
                int h = Math.min(charHeight-y,
                                 (clipRect.y+clipRect.height+cellHeight-1)/
                                 cellHeight-y);
                repaintArea(graphics,x,y,w,h);
                repaintCursorInRectangle(graphics,x,y,w,h);
            }
            if (useDoubleBuffering)
                g.drawImage(consoleImage,0,0,width,height,this);
            // Check whether we repainted the entire invalidated region
            if (isInvalid &&
                invalidX*cellWidth >= clipRect.x &&
                (invalidX+invalidW)*cellWidth <= clipRect.x + clipRect.width &&
                (charHeight*cellHeight <= clipRect.height ||
                 ((invalidY-scrollOffset)*cellHeight >= clipRect.y &&
                  (invalidY-scrollOffset+invalidH)*cellHeight <=
                                                clipRect.y + clipRect.height)))
                isInvalid = false;
        }
        synchronized (repaintDoneMonitor) {
            repaintDoneMonitor.notify();
        }
        return;
    }


    public void update(Graphics g) {
        // Subclass update so that we don't unneccessarily fill with the
        // background color (avoids flashing)
        paint(g);
        return;
    }


    public void reshape(int x,int y,int w,int h) {
        super.reshape(x,y,w,h);
        // Reshape currently does not implement much of a useful policy.
        // It will always reinitialize the console if the size changed.
        // This is fine, because the applet doesn't support size changes
        // anyway. If this ever changes, we probably want to implement
        // an algorithm that tries to conserve console contents whereever
        // possible.
        if (width == w && height == h)
            return;
        Container parent  = getParent();
        if (parent == null)
            return;
        Graphics graphics = parent.getGraphics();
        if (graphics == null)
            return;
        resetConsole(graphics,w,h);
        return;
    }


    public boolean isDoubleBuffered() {
        return(useDoubleBuffering);
    }


    public boolean gotFocus(Event ev,Object what) {
        if (!isFocused) {
            isFocused = true;
            invalidateRect(cursorX,cursorY,1,1);
            if (sizeIsValid)
                flushOperations();
        }
        return(super.gotFocus(ev,what));
    }


    public boolean lostFocus(Event ev,Object what) {
        if (isFocused) {
            isFocused = false;
            invalidateRect(cursorX,cursorY,1,1);
            if (sizeIsValid)
                flushOperations();
        }
        return(super.lostFocus(ev,what));
    }


    public void setBellType(boolean visualBell) {
        this.visualBell = visualBell;
        return;
    }


    public void beep() {
        if (visualBell) {
            invertScreen(!isInverted);
            invertScreen(!isInverted);
        } else {
            try {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                toolkit.beep();
            } catch (Throwable e) {
                // JDK 1.0.2
            }
        }
        return;
    }

    
    public void displayStatus(String status) {
        return;
    }


    public void scrollBack() {
        setScrollOffset(Math.max(0,scrollOffset-charHeight));
        scrollbar.setValue(scrollOffset);
        return;
    }


    public void scrollFore() {
        setScrollOffset(Math.min(scrollBackLines,scrollOffset+charHeight));
        scrollbar.setValue(scrollOffset);
        return;
    }


    public void clearScrollBackBuffer() {
        synchronized (this) {
            scrollbar.setValues(0,charHeight,0,charHeight);
            System.arraycopy(consoleCharBuffer,scrollBackLines,
                             consoleCharBuffer,0,charHeight);
            System.arraycopy(consoleAttrBuffer,scrollBackLines,
                             consoleAttrBuffer,0,charHeight);
            invalidY       -= scrollBackLines;
            scrollBackLines = 0;
            scrollOffset    = 0;
            invalidateRect(0,0,charWidth,charHeight);
            flushOperations();
        }
        return;
    }


    protected void resetConsole(Graphics g,int width,int height) {
        this.width              = width;
        this.height             = height;
        sizeIsValid             = true;
        isFocused               = false;
        Graphics graphics;
        if (useDoubleBuffering) {
            consoleImage        = createImage(width,height);
            try {
                consoleImage.flush();
                consoleGraphics.dispose();
            } catch (NullPointerException e) {
                // ignore
            }
            consoleGraphics     = consoleImage.getGraphics();
            graphics            = consoleGraphics;
        } else
            graphics            = g;
        cursorX                 = 0;
        cursorY                 = 0;
        isInverted              = false;
        isAlternateScreen       = false;
        utfEnabled              = false;
        settingClipboard        = false;
        isSelected              = false;
        Dimension fontSize      = computeFontSize(graphics);
        cellWidth               = fontSize.width;
        cellHeight              = fontSize.height;
        charWidth               = width/cellWidth;
        charHeight              = height/cellHeight;
        if (charWidth <= 0)
            charWidth = 1;
        if (charHeight <= 0)
            charHeight = 1;
        scrollbar.setValues(0,charHeight,0,charHeight);
        scrollbar.setLineIncrement(1);
        scrollbar.setPageIncrement(charHeight);
        scrollBackLines         = 0;
        scrollOffset            = 0;
        blanks                  = new char[charWidth];
        consoleCharBuffer       = new char[charHeight+maxScrollBackLines][];
        consoleAttrBuffer       = new char[charHeight+maxScrollBackLines][];
        alternateCharBuffer     = new char[charHeight][];
        alternateAttrBuffer     = new char[charHeight][];
        consoleAttrBuffer[0]    = new char[charWidth];
        for (int x = charWidth; x-- > 0; ) {
            blanks[x]               = '\000';
            consoleAttrBuffer[0][x] = 0xF0; // black on white
        }
        for (int y = charHeight; y-- > 0; ) {
            consoleCharBuffer[y] = new char[charWidth];
            System.arraycopy(blanks,0,consoleCharBuffer[y],0,charWidth);
            if (y > 0) {
                consoleAttrBuffer[y] = new char[charWidth];
                System.arraycopy(consoleAttrBuffer[0],0,
                                 consoleAttrBuffer[y],0,charWidth);
            }
            alternateCharBuffer[y] = new char[charWidth];
            System.arraycopy(blanks,0,alternateCharBuffer[y],0,charWidth);
            alternateAttrBuffer[y] = new char[charWidth];
            System.arraycopy(consoleAttrBuffer[0],0,
                             alternateAttrBuffer[y],0,charWidth);
        }
        if (emulator != null)
            emulator.setSize(new Dimension(charWidth,charHeight));
        return;
    }


    private Rectangle hitClip(Rectangle clip,int x,int y,int w,int h) {
        w = Math.min(clip.x+clip.width, x+w);
        h = Math.min(clip.y+clip.height,y+h);
        x = Math.max(clip.x,            x  );
        y = Math.max(clip.y,            y  );
        w-= x;
        h-= y;
        if (w > 0 && h > 0) {
            return(new Rectangle(x,y,w,h));
        } else {
            return(null);
        }
    }

    
    public void setCursorAttr(int setAttr,int xorAttr) {
        if (setAttr == 0 && xorAttr == 0) {
            setAttr = CURSOR_SET_ATTR;
            xorAttr = CURSOR_XOR_ATTR;
        }
        // The Linux console uses hardware colors to specify cursor attributes,
        // whereas it uses VT100 colors everywhere else. This looks like a bug
        // in the Linux code but I decided to copy the behavior since other
        // programs rely on it.
        cursorSetAttr = (char)((vgaColors[(setAttr >> 4) & 0x0F] << 4) |
                                vgaColors[ setAttr       & 0x0F]      );
        cursorXORAttr = (char)((vgaColors[(xorAttr >> 4) & 0x0F] << 4) |
                                vgaColors[ xorAttr       & 0x0F]      );
        return;
    }


    Color setAttrColors(Graphics graphics,char attr) {
        // This code has a couple of rather strange properties, but I
        // decided to implement it this way because it closely models the
        // behavior that the Linux console exhibits when driving VGA
        // cards. This is based on the assumption that many programs are
        // written to properly display on the Linux console even if its
        // behavior is not always obvious. N.B. this is different from
        // xterm behavior which doesn't support the "dim" and "blink"
        // attributes!
        int bg = (attr >> 4) & 0xF;
        int fg =  attr       & 0xF;
        if ((attr & REVERSE) != 0) {
            int tmp = bg & 0x7;
            bg = (bg & 0x8) | (fg & 0x7);
            fg = (fg & 0x8) | tmp;
            if ((attr & (REVERSE|DIM)) == (REVERSE|DIM)) {
                bg  = 0; // black
                fg |= 8;
            }
        }
        if ((attr & (REVERSE|DIM)) == DIM)
            fg  = 8; // dark grey
        else if ((attr & BRIGHT) != 0)
            fg |= 8;
        if ((attr & BLINK) != 0)
            bg ^= 0x8;
        Color background;
        Color foreground;
        if (isInverted) {
            background = consoleColors[fg];
            foreground = consoleColors[bg];
        } else {
            background = consoleColors[bg];
            foreground = consoleColors[fg];
        }
        if ((attr & SELECTED) != 0) {
            if (background == Color.black)
                background = Color.darkGray;
            else {
                float[] hsb = Color.RGBtoHSB(background.getRed(),
                                             background.getGreen(),
                                             background.getBlue(),null);
                background = Color.getHSBColor(hsb[0],hsb[1],
                                               (float)(hsb[2] * 0.5));
            }
        }
        graphics.setColor(background);
        return(foreground);
    }

    
    void setScrollOffset(int scrollOffset) {
        synchronized (this) {
            if (scrollOffset < 0)
                scrollOffset = 0;
            else if (scrollOffset > scrollBackLines)
                scrollOffset = scrollBackLines;
            if (scrollOffset == this.scrollOffset)
                return;
            this.scrollOffset = scrollOffset;
            invalidateRect(0,-scrollBackLines,
                           charWidth,scrollBackLines+charHeight);
            flushOperations();
        }
        return;
    }


    private void invalidateRect(int x,int y,int w,int h) {
        synchronized (this) {
            y += scrollBackLines;
            if (isInvalid) {
                if (x < invalidX) {
                    invalidW += invalidX - x;
                    invalidX  = x; }
                if (x+w > invalidX + invalidW)
                    invalidW = x+w-invalidX;
                if (y < invalidY) {
                    invalidH += invalidY - y;
                    invalidY  = y; }
                if (y+h > invalidY + invalidH)
                    invalidH = y+h-invalidY;
            } else {
                isInvalid = true;
                invalidX  = x;
                invalidY  = y;
                invalidW  = w;
                invalidH  = h;
            }
            if (invalidX < 0) {
                invalidW += invalidX;
                invalidX  = 0;
            }
            if (invalidX + invalidW > charWidth)
                invalidW = charWidth - invalidX;
            if (invalidY < 0) {
                invalidH += invalidY;
                invalidY  = 0;
            }
            if (invalidY + invalidH > charHeight + scrollBackLines)
                invalidH = charHeight + scrollBackLines - invalidH;
            if (invalidW <= 0 || invalidH <= 0)
                isInvalid = false;
        }
        return;
    }


    private void updateOperationsCounter(int num,int weight) {
        operationsCounter += num*weight;
        // Force redraws every once in a while to improve interactivity
        if (operationsCounter >= OPERATIONS_THRESHOLD)
            flushOperations();
        return;
    }


    public void flushOperations() {
        int x,y,w,h;
        operationsCounter = 0;
        synchronized (this) {
            if (!isInvalid)
                return;
            x = invalidX                  * cellWidth;
            y = (invalidY - scrollOffset) * cellHeight;
            h = invalidH                  * cellHeight;
            if (invalidX + invalidW >= charWidth)
                w = width;
            else
                w = invalidW              * cellWidth;
        }
        // Every so often, force an immediate repaint
        if (flushCounter++ >= FLUSH_THRESHOLD) {
            flushCounter = 0;
            synchronized (repaintMonitor) {
                repaint(x,y,w,h);
                try {
                    repaintMonitor.wait(1);
                } catch (InterruptedException e) {
                }
            }
        } else
            repaint(50,x,y,w,h);
        return;
    }
            

    public void setCursorVisibility(boolean visible) {
        cursorVisible = visible;
        invalidateRect(cursorX,cursorY,1,1);
        return;
    }


    public boolean getCursorVisibility() {
        return(cursorVisible);
    }


    public void setCursorPosition(Point position) {
        updateOperationsCounter(1,CURSORPOS_WEIGHT);
        if (cursorX == position.x &&
            cursorY == position.y)
            return;
        if (cursorVisible) {
            cursorVisible = false;
            invalidateRect(cursorX,cursorY,1,1);
            cursorVisible = true;
            // If the cursor jumped too far, flush operations stack now; thus
            // we can avoid invalidating a very large rectangle
            if ((Math.abs(cursorX - position.x) + 1)*
                (Math.abs(cursorY - position.y) + 1) > 200)
                flushOperations();
        }
        if (position.x < 0)
            cursorX = 0;
        else if (position.x >= charWidth)
            cursorX = charWidth - 1;
        else
            cursorX = position.x;
        if (position.y < 0)
            cursorY = 0;
        else if (position.y >= charHeight)
            cursorY = charHeight - 1;
        else
            cursorY = position.y;
        if (cursorVisible)
            invalidateRect(cursorX,cursorY,1,1);
        return;
    }


    public void drawString(String s,int x,int y,char attr) {
        int length = s.length();
        updateOperationsCounter(length,DRAWSTRING_WEIGHT);
        if (x < 0 || x >= charWidth ||
            y < 0 || y >= charHeight)
            return;
        refreshSelection(false);
        length = Math.min(length,charWidth-x);
        s.getChars(0,length,consoleCharBuffer[y+scrollBackLines],x);
        for (int i = 0; i < length; i++ )
            consoleAttrBuffer[y+scrollBackLines][x+i] = attr;
        invalidateRect(x,y,length,1);
        return;
    }


    public void scrollRegion(int x,int y,int w,int h,int incX,int incY,
                             char attr) {
        boolean oldCursorVisible = cursorVisible;
        if (oldCursorVisible)
            setCursorVisibility(false);
        boolean dontScroll = false;
        int left  = (incX < 0) ? -incX : 0;
        int right = (incX > 0) ?  incX : 0;
        int up    = (incY < 0) ? -incY : 0;
        int down  = (incY > 0) ?  incY : 0;
        // Clip region against screen size
        w += x;
        if (x < left)
            x = left;
        if (w > charWidth-right)
            w = charWidth-right;
        if ((w -= x) <= 0)
            dontScroll = true;
        h += y;
        if (y < up)
            y = up;
        if (h > charHeight-down)
            h = charHeight-down;
        if ((h -= y) < 0)
            dontScroll = true;
        // Update buffers
        synchronized (this) {
            if (!dontScroll) {
                int yS = y + scrollBackLines;
                if (incX == 0 && x == 0 && w == charWidth) {
                    // Scrolling entire lines
                    if (incY < 0) {
                        // Scrolling up
                        if (!isAlternateScreen &&
                            y == -incY && h == charHeight+incY) {
                            // Scrolling up with adding to the scrollback buf.
                            int adjust=scrollBackLines-maxScrollBackLines-incY;
                            if (adjust > 0) {
                                // Need to scroll the buffer
                                char[][] tmp = new char[adjust][];
                                System.arraycopy(consoleCharBuffer,0,
                                                 tmp,0,adjust);
                                System.arraycopy(consoleCharBuffer,adjust,
                                                 consoleCharBuffer,0,
                                                 yS+h-adjust);
                                System.arraycopy(tmp,0,consoleCharBuffer,
                                                 yS+h-adjust,adjust);
                                System.arraycopy(consoleAttrBuffer,0,
                                                 tmp,0,adjust);
                                System.arraycopy(consoleAttrBuffer,adjust,
                                                 consoleAttrBuffer,0,
                                                 yS+h-adjust);
                                System.arraycopy(tmp,0,consoleAttrBuffer,
                                                 yS+h-adjust,adjust);
                                for (int Y = yS+h;
                                     Y < maxScrollBackLines+charHeight; Y++) {
                                    consoleCharBuffer[Y] = new char[charWidth];
                                    consoleAttrBuffer[Y] = new char[charWidth];
                                }
                                if (scrollOffset == scrollBackLines)
                                    scrollOffset = maxScrollBackLines;
                                else {
                                    scrollOffset -= adjust;
                                    if (scrollOffset < 0)
                                        scrollOffset = 0;
                                }
                                scrollBackLines = maxScrollBackLines;
                                invalidateRect(0,-scrollBackLines,charWidth,
                                               maxScrollBackLines+charHeight);
                            } else {
                                // The buffer is not full yet, we can get by
                                // with adjusting the offset
                                if (scrollOffset == scrollBackLines)
                                    scrollOffset -= incY;
                                scrollBackLines -= incY;
                                for (int Y = yS+h; Y < yS+h-incY; Y++) {
                                    consoleCharBuffer[Y] = new char[charWidth];
                                    consoleAttrBuffer[Y] = new char[charWidth];
                                }
                            }
                            scrollbar.setValues(scrollOffset,charHeight,0,
                                                scrollBackLines + charHeight);
                        } else {
                            // Scrolling up without adding to the scrollback
                            // buffer
                            char[][] tmp = new char[-incY][];
                            System.arraycopy(consoleCharBuffer,yS+incY,
                                             tmp,0,-incY);
                            System.arraycopy(consoleCharBuffer,yS,
                                             consoleCharBuffer,yS+incY,h);
                            System.arraycopy(tmp,0,consoleCharBuffer,yS+h+incY,
                                             -incY);
                            System.arraycopy(consoleAttrBuffer,yS+incY,
                                             tmp,0,-incY);
                            System.arraycopy(consoleAttrBuffer,yS,
                                             consoleAttrBuffer,yS+incY,h);
                            System.arraycopy(tmp,0,consoleAttrBuffer,yS+h+incY,
                                             -incY);
                        }
                    } else {
                        // Scrolling down
                        char[][] tmp = new char[incY][];
                        System.arraycopy(consoleCharBuffer,yS+h,tmp,0,incY);
                        System.arraycopy(consoleCharBuffer,yS,
                                         consoleCharBuffer,yS+incY,h);
                        System.arraycopy(tmp,0,consoleCharBuffer,yS,incY);
                        System.arraycopy(consoleAttrBuffer,yS+h,tmp,0,incY);
                        System.arraycopy(consoleAttrBuffer,yS,
                                         consoleAttrBuffer,yS+incY,h);
                        System.arraycopy(tmp,0,consoleAttrBuffer,yS,incY);
                    }
                } else
                    // Scrolling partial lines
                    if (incY < 0)
                        // Scrolling up or horizontally within a line
                        for (int Y = yS; Y < yS+h; Y++) {
                            System.arraycopy(consoleCharBuffer[Y],x,
                                             consoleCharBuffer[Y+incY],x+incX,
                                             w);
                            System.arraycopy(consoleAttrBuffer[Y],x,
                                             consoleAttrBuffer[Y+incY],x+incX,
                                             w);
                        }
                    else
                        // Scrolling down
                        for (int Y = yS+h; Y-- > yS; ) {
                            System.arraycopy(consoleCharBuffer[Y],x,
                                             consoleCharBuffer[Y+incY],x+incX,
                                             w);
                            System.arraycopy(consoleAttrBuffer[Y],x,
                                             consoleAttrBuffer[Y+incY],x+incX,
                                             w);
                        }
                invalidateRect(x+incX,y+incY,w,h);
            }
            // Different terminal emulators disagree on the attributes that
            // are used for scrolling. The consensus seems to be, never to
            // fill with underlined spaces. N.B. this is different from the
            // cases when the user blanks a region. User-initiated blanking
            // always fills with all of the current attributes.
            attr &= ~UNDERLINE;
            // Clear blank regions
            if (incX > 0)
                clearRegion(x,y,incX,h,attr);
            else if (incX < 0)
                clearRegion(x+w+incX,y,-incX,h,attr);
            if (incY > 0)
                clearRegion(x,y,w,incY,attr);
            else if (incY < 0)
                clearRegion(x,y+h+incY,w,-incY,attr);
        }
        // Redraw cursor
        if (oldCursorVisible)
            setCursorVisibility(true);
        updateOperationsCounter(1,SCROLL_WEIGHT);
        return;
    }


    public void clearRegion(int x,int y,int w,int h,char attr) {
        // Clip region against screen size
        w += x;
        if (x < 0)
            x = 0;
        if (w > charWidth)
            w = charWidth;
        if ((w -= x) <= 0)
            return;
        h += y;
        if (y < 0)
            y = 0;
        if (h > charHeight)
            h = charHeight;
        if ((h -= y) <= 0)
            return;
        // Update buffers
        refreshSelection(false);
        for (int X = x+w; X-- > x; ) {
            consoleAttrBuffer[y+scrollBackLines][X] = attr;
        }
        for (int Y = y+h+scrollBackLines; --Y > y+scrollBackLines; ) {
            System.arraycopy(blanks,0,consoleCharBuffer[Y],x,w);
            System.arraycopy(consoleAttrBuffer[y+scrollBackLines],x,
                             consoleAttrBuffer[Y],x,w);
        }
        System.arraycopy(blanks,0,consoleCharBuffer[y+scrollBackLines],x,w);
        // Clear region in offscreen image
        invalidateRect(x,y,w,h);
        updateOperationsCounter(1,CLEAR_WEIGHT);
        return;
    }


    public void invertScreen(boolean state) {
        if (isInverted == state)
            return;
        isInverted = state;
        invalidateRect(0,-scrollBackLines,charWidth,
                       scrollBackLines+charHeight);
        synchronized (repaintDoneMonitor) {
            repaint(0,0,width,charHeight*cellHeight);
            try {
                repaintDoneMonitor.wait(1000);
            } catch (InterruptedException e) {
            }
        }
        return;
    }


    public void setUTFEnabled(boolean state) {
        utfEnabled = state;
        return;
    }


    public void enableAlternateScreen(boolean state) {
        if (state == isAlternateScreen)
            return;
        refreshSelection(false);
        isAlternateScreen = state;
        for (int y = charHeight; y-- > 0; ) {
            char[] tmp;
            tmp = alternateCharBuffer[y];
            alternateCharBuffer[y] = consoleCharBuffer[y + scrollBackLines];
            consoleCharBuffer[y + scrollBackLines] = tmp;
            tmp = alternateAttrBuffer[y];
            alternateAttrBuffer[y] = consoleAttrBuffer[y + scrollBackLines];
            consoleAttrBuffer[y + scrollBackLines] = tmp;
        }
        invalidateRect(0,0,charWidth,charHeight);
        return;
    }


    private final void refreshSelection(boolean state) {
        synchronized (this) {
            if (!state && !isSelected)
                return;
            int x1,y1,x2,y2;
            if (selectedY1 < selectedY2 ||
                (selectedY1 == selectedY2 && selectedX1 < selectedX2)) {
                x1 = selectedX1; x2 = selectedX2;
                y1 = selectedY1; y2 = selectedY2;
            } else {
                x1 = selectedX2; x2 = selectedX1;
                y1 = selectedY2; y2 = selectedY1;
            }
            int xStart = x1;
            for (int y = y1; y <= y2; y++) {
                int xStop = ((y == y2) ? x2+1 : charWidth);
                while(--xStop >= xStart &&
                      consoleCharBuffer[y][xStop] == '\000');
                for (; xStop >= xStart; xStop--)
                    if (state)
                        consoleAttrBuffer[y][xStop] |= SELECTED;
                    else
                        consoleAttrBuffer[y][xStop] &=~SELECTED;
                xStart = 0;
            }
            invalidateRect(0,y1-scrollBackLines,charWidth,y2-y1+1);
            isSelected = state;
        }
        return;
    }


    private final String getSelection() {
        if (!isSelected)
            return(null);
        int x1,y1,x2,y2;
        if (selectedY1 < selectedY2 ||
            (selectedY1 == selectedY2 && selectedX1 < selectedX2)) {
            x1 = selectedX1; x2 = selectedX2;
            y1 = selectedY1; y2 = selectedY2;
        } else {
            x1 = selectedX2; x2 = selectedX1;
            y1 = selectedY2; y2 = selectedY1;
        }
        StringBuffer selection = new StringBuffer();
        int          xStart    = x1;
        for (int y = y1; y <= y2; y++) {
            int xStop = ((y == y2) ? x2+1 : charWidth);
            while(--xStop >= xStart && consoleCharBuffer[y][xStop] == '\000');
            if (xStop >= xStart) {
                char[] tmp = new char[xStop-xStart+1];
                System.arraycopy(consoleCharBuffer[y],xStart,tmp,0,tmp.length);
                for (int i = tmp.length; i-- > 0; )
                    if (tmp[i] == '\000')
                        tmp[i] = ' ';
                selection.append(tmp);
                if (y != y2)
                    selection.append("\n");
            }
            xStart = 0;
        }
        return(selection.toString());
    }


    static String toUTF8(String src) {
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < src.length(); i++) {
            char ch = src.charAt(i);
            if (ch <= '\177')
                buffer.append(ch);
            else if (ch <= '\u07FF') {
                buffer.append((char)(0xC0 | (ch >> 6)));
                buffer.append((char)(0x80 | (ch & 0x3F)));
            } else {
                buffer.append((char)(0xE0 | (ch >> 12)));
                buffer.append((char)(0x80 | ((ch >> 6) & 0x3F)));
                buffer.append((char)(0x80 | (ch & 0x3F)));
            }
        }
        return(buffer.toString());
    }


    public void lostOwnership(Clipboard clipboard,String contents) {
        if (!settingClipboard) {
            refreshSelection(false);
            flushOperations();
        }
        return;
    }


    public String processMouseEvent(int event,int x,int y,int clickCount,
                                    int button,int modifiers) {
        synchronized (this) {
            int yScroll = y - scrollBackLines;
            int yScreen = y - scrollOffset;
            if (!isSelected && emulator != null &&
                x       >= 0 && x       < charWidth &&
                yScroll >= 0 && yScroll < charHeight) {
                String rc = emulator.processMouseEvent(event,x,yScroll,
                                                  clickCount,button,modifiers);
                if (rc != null)
                    return(rc);
            }
            switch (event) {
            case Event.MOUSE_UP:
                if ((button & 1) != 0 && isSelected) {
                    String selection = getSelection();
                    settingClipboard = true;
                    clipboard.setContents(selection,this);
                    settingClipboard = false;
                    return("");
                } else if ((button & 2) != 0) {
                    String selection = clipboard.getContents(this);
                    if (selection == null)
                        return(null);
                    if (utfEnabled)
                        selection = toUTF8(selection);
                    StringBuffer buffer = new StringBuffer();
                    for (int i = 0; i < selection.length(); i++) {
                        char ch = selection.charAt(i);
                        if (ch != '\n' && ch != '\r' && ch != '\t' &&
                            ch < ' ' || ch > '\377') {
                            buffer.append(' ');
                        } else {
                            buffer.append(ch);
                        }
                    }
                    return(buffer.toString());
                }
                return(null);
            case Event.MOUSE_DOWN:
                if ((button & 1) != 0 &&
                    x       >= 0 && x       < charWidth &&
                    yScreen >= 0 && yScreen < charHeight) {
                    refreshSelection(false);
                    flushOperations();
                    selectedX1 = x;
                    selectedY1 = y;
                    selectedX2 = x;
                    selectedY2 = y;
                }
                return(null);
            case Event.MOUSE_DRAG:
                if ((button & 1) == 0)
                    return(null);
                refreshSelection(false);
                if (x < 0)
                    selectedX2 = 0;
                else if (x >= charWidth)
                    selectedX2 = charWidth-1;
                else
                    selectedX2 = x;
                if (yScreen < 0)
                    if (scrollOffset <= 0)
                        selectedY2 = 0;
                    else
                        selectedY2 = scrollOffset-1;
                else if (yScreen >= charHeight)
                    if (scrollOffset >= scrollBackLines)
                        selectedY2 = scrollBackLines + charHeight - 1;
                    else
                        selectedY2 = scrollOffset + charHeight;
                else
                    selectedY2 = y;
                refreshSelection(true);
                if (scrollMonitor == null)
                    new ScrollMonitor();
                flushOperations();
                return("");
            case Event.MOUSE_EXIT:
            case Event.MOUSE_ENTER:
            case Event.MOUSE_MOVE:
            default:
                return(null);
            }
        }
    }


    public class ScrollMonitor implements Runnable {

        public ScrollMonitor() {
            synchronized (Console.this) {
                if (scrollMonitor == null) {
                    scrollMonitor = this;
                    Thread thread = new Thread(this);
                    thread.start();
                }
            }
            return;
        }
        

        public void run() {
                for (;;) {
                    synchronized (Console.this) {
                        if (!isSelected) {
                            scrollMonitor = null;
                            return;
                        }
                        if (selectedY2 < scrollOffset &&
                            scrollOffset > 0) {
                            if (scrollOffset > 1) {
                                refreshSelection(false);
                                selectedY2 = scrollOffset - 2;
                                refreshSelection(true);
                            }
                            scrollbar.setValue(scrollOffset - 1);
                            setScrollOffset(scrollOffset - 1);
                        } else if (selectedY2 >= scrollOffset+charHeight &&
                                   scrollOffset < scrollBackLines) {
                            if (scrollOffset < scrollBackLines-1) {
                                refreshSelection(false);
                                selectedY2 = scrollOffset+charHeight + 1;
                                refreshSelection(true);
                            }
                            scrollbar.setValue(scrollOffset + 1);
                            setScrollOffset(scrollOffset + 1);
                        }
                    }
                    try {
                        Thread.sleep(80);
                    } catch (InterruptedException e) {
                        // don't care
                    }
                }
        }
    }


    public class ConsoleScrollbar extends Scrollbar {

        private boolean hasBug;


        public ConsoleScrollbar(int orientation) {
            super(orientation);
            String vendor  = System.getProperty("java.vendor");
            String version = System.getProperty("java.version");
            hasBug = vendor.equals("Netscape Communications Corporation") &&
                     version.startsWith("1.0");
            return;
        }


        public boolean handleEvent(Event ev) {
            switch (ev.id) {
            case Event.KEY_PRESS:
            case Event.KEY_ACTION:
                if (isFocused)
                    return(Console.this.handleEvent(ev));
                break;
            case Event.SCROLL_ABSOLUTE:
            case Event.SCROLL_LINE_UP:
            case Event.SCROLL_LINE_DOWN:
            case Event.SCROLL_PAGE_UP:
            case Event.SCROLL_PAGE_DOWN:
                setScrollOffset(((Integer)ev.arg).intValue());
                return(true);
            default:
                break;
            }
            return(super.handleEvent(ev));
        }


        public void setValues(int value,int visible,int minimum,int maximum) {
            super.setValues(value,visible,minimum,
                            hasBug ? maximum-visible : maximum);
            return;
        }
    }
}

/*
 * Local variables:
 * mode: java
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
