package com.shellinabox.console;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;


/**
 * Loadable user supplied fonts with support for all Unicode characters.
 *
 * Raster operations while being the traditional way to display loadable
 * fonts are very inefficient in current Java implementations. Moreover,
 * the use of a large amount of raster operations within a small period of
 * time can cause Netscape/X11 to lock up by saturating the the X11 command
 * queue. However, graphics primitives such as drawLine and fillRectangle
 * have proven to be highly efficient. Thus, we encode all glyph information
 * as a set of drawing primitives. The utility "unicodeglyphs" computes an
 * optimal set for any given glyph. It takes multiple Linux-style *.psfu
 * files as input and outputs a new font file. The file format supports up to
 * 65536 distinct character positions. This is sufficient to encode the
 * entire range of Unicode characters. While UTF-8 supports up to 2^32
 * distinct codes, there is currently no need to support this extended range.
 */
public class VectorFontConsole extends Console {

    private static final char    UNKNOWN_GLYPH  = '\uFFFD';

    private VectorFont vectorFont;
    private Hashtable  fontData;
    private int[][]    indexedFontData;
    private int        baseLine;
    private int        numGlyphs;
    private int[]      unknownGlyph;


    public VectorFontConsole(int scrollBackLines,boolean visualBell,
                             URL fontURL) {
        super(scrollBackLines,visualBell);
        vectorFont      = VectorFont.getSharedFont(fontURL);
        fontData        = vectorFont.getFontData();
        indexedFontData = vectorFont.getIndexedFontData();
        baseLine        = vectorFont.getBaseLine();
        numGlyphs       = vectorFont.getNumGlyphs();
        unknownGlyph    = vectorFont.getUnknownGlyph();
        return;
    }


    void repaintArea(Graphics graphics,int x,int y,int w,int h) {
        // Redraw text
        StringBuffer lineBuf = new StringBuffer(w);
        for (int Y = y; Y < y+h; Y++) {
            lineBuf.setLength(0);
            int  lineLength   = 0;
            int  X1           = x;
            char[] characters = consoleCharBuffer[Y+scrollOffset];
            char[] attributes = consoleAttrBuffer[Y+scrollOffset];
            char attr         = attributes[X1];
            int  offY         = Y*cellHeight;
            for (int X = x; X <= x+w; X++) {
                if (X == x+w || attributes[X] != attr) {
                    Color foreGround = setAttrColors(graphics,attr);
                    int offX = X1*cellWidth;
                    graphics.fillRect(offX,offY,lineLength*cellWidth,
                                      cellHeight);
                    graphics.setColor(foreGround);
                    for (int i = 0; i < lineLength; i++, offX += cellWidth) {
                        char ch = lineBuf.charAt(i);
                        if (ch == '\000' || ch == ' ')
                            continue;
                        int[] coords;
                        if (ch < '\u0100')
                            coords = indexedFontData[ch];
                        else
                            coords = (int[])fontData.get(new Character(ch));
                        if (coords == null)
                            coords = unknownGlyph;
                        for (int idx = 0; idx < coords.length; ) {
                            int x1 = coords[idx++] + offX;
                            int y1 = coords[idx++] + offY;
                            int x2 = coords[idx++];
                            int y2 = coords[idx++];
                            if (x2 >= 100)
                                graphics.fillRect(x1,y1,x2-100,y2);
                            else
                                graphics.drawLine(x1,y1,x2+offX,y2+offY);
                        }
                    }
                    if ((attr & UNDERLINE) != 0) {
                        graphics.drawLine(X1*cellWidth,offY+baseLine,
                                          offX-1,offY+baseLine);
                    }
                    if (X == x+w)
                        break;
                    lineBuf.setLength(0);
                    lineLength = 0;
                    X1         = X;
                    attr       = attributes[X1];
                }
                lineBuf.append(characters[X]);
                lineLength++;
            }
        }
        repaintCursorInRectangle(graphics,x,y,w,h);
        return;
    }


    void repaintCursorInRectangle(Graphics graphics,int x,int y,int w,int h) {
        // Redraw cursor
        int cursorYB = cursorY  + scrollBackLines;
        int cursorYS = cursorYB - scrollOffset;
        if (cursorVisible &&
            cursorX >= x && cursorX < x+w &&
            cursorYS>= y && cursorYS< y+h) {
            char attr = consoleAttrBuffer[cursorYB][cursorX];
            char cursorAttr = (char)((attr | cursorSetAttr) ^ cursorXORAttr);
            char ch = consoleCharBuffer[cursorYB][cursorX];
            Color foreGround = setAttrColors(graphics,
                                             isFocused ? cursorAttr : attr);
            graphics.fillRect(cursorX*cellWidth,cursorYS*cellHeight,
                              cellWidth,cellHeight);
            if (!isFocused) {
                setAttrColors(graphics,cursorAttr);
                graphics.drawRect(cursorX*cellWidth,cursorYS*cellHeight,
                                  cellWidth-1,cellHeight-1);
            }
            int offX = cursorX*cellWidth;
            int offY = cursorYS*cellHeight;
            if (ch != '\000' && ch != ' ') {
                int[] coords;
                if (ch < '\u0100')
                    coords = indexedFontData[ch];
                else
                    coords = (int[])fontData.get(new Character(ch));
                if (coords == null)
                    coords = unknownGlyph;
                graphics.setColor(foreGround);
                for (int idx = 0; idx < coords.length; ) {
                    int x1 = coords[idx++] + offX;
                    int y1 = coords[idx++] + offY;
                    int x2 = coords[idx++];
                    int y2 = coords[idx++];
                    if (x2 >= 100)
                        graphics.fillRect(x1,y1,x2-100,y2);
                    else
                        graphics.drawLine(x1,y1,x2+offX,y2+offY);
                }
            }
            if ((attr & UNDERLINE) != 0) {
                graphics.drawLine(offX,offY+baseLine,
                                  offX+cellWidth-1,offY+baseLine);
            }

        }
        return;
    }


    Dimension computeFontSize(Graphics graphics) {
        int width  = vectorFont.getCellWidth();
        int height = vectorFont.getCellHeight();
        return(new Dimension(width,height));
    }


    protected void resetConsole(Graphics g,int width,int height) {
        super.resetConsole(g,width,height);
        for (Enumeration enum = fontData.keys(); enum.hasMoreElements(); ) {
            Character ch = (Character)enum.nextElement();
            int x = (int)(charWidth*Math.random());
            int y = (int)(charHeight*Math.random());
            consoleCharBuffer[y+scrollOffset][x] = ch.charValue();
            consoleAttrBuffer[y+scrollOffset][x] = 0xF7;
        }
        return;
    }

}

/*
 * Local variables:
 * mode: java
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
