package com.shellinabox.emulation;

import com.shellinabox.console.Console;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Point;


/**
 * The VT102 emulator is a port of the Linux console driver to Java. Some of
 * the very hardware-specific features of the Linux driver have not been
 * copied but any other differences between the two emulators should be
 * considered a bug and reported to the author.
 *
 * There is built-in support for XTerm-style mouse position reporting,
 * alternate screen management, and status bars. There is support for
 * Screen-style back tabs, and visual bell. ANSI-style parametrized scrolling
 * and tabs have been added.
 *
 * Most terminal emulators that pretend to emulate the DEC VT102 are very
 * similar and terminfo/termcap entries for "vt102", "xterm", "screen",
 * "linux", "ansi", ... can be used almost interchangably. The only major
 * difference is in the response to pressing function keys. There does not
 * appear to be a widely accepted standard. This emulator can be configured
 * to return suitable codes for Linux, ANSI, VT102, and XTerm.
 */
public class VT102 implements UnicodeMap {

    public  static final int    LINUX         =  0;
    public  static final int    VT102         =  1;
    public  static final int    ANSI          =  2;
    public  static final int    XTERM         =  3;
    public  static final int    XTERM_COLOR   =  4;

    private static final int    NPAR          = 16;
    private static final char   UNKNOWN_GLYPH = '\uFFFD';
    private static final String VT102ID       = "\u001B[?6c";

    private static final int    ESnormal      =  0;
    private static final int    ESesc         =  1;
    private static final int    ESsquare      =  2;
    private static final int    ESgetpars     =  3;
    private static final int    ESgotpars     =  4;
    private static final int    ESfunckey     =  5;
    private static final int    EShash        =  6;
    private static final int    ESsetG0       =  7;
    private static final int    ESsetG1       =  8;
    private static final int    ESpercent     =  9;
    private static final int    ESignore      = 10;
    private static final int    ESnonstd      = 11;
    private static final int    ESpalette     = 12;
    private static final int    ESstatus      = 13;

    private static final int    HOME          =  0;
    private static final int    END           =  1;
    private static final int    PGUP          =  2;
    private static final int    PGDN          =  3;
    private static final int    UP            =  4;
    private static final int    DOWN          =  5;
    private static final int    LEFT          =  6;
    private static final int    RIGHT         =  7;
    private static final int    F1            =  8;
    private static final int    F2            =  9;
    private static final int    F3            = 10;
    private static final int    F4            = 11;
    private static final int    F5            = 12;
    private static final int    F6            = 13;
    private static final int    F7            = 14;
    private static final int    F8            = 15;
    private static final int    F9            = 16;
    private static final int    F10           = 17;
    private static final int    F11           = 18;
    private static final int    F12           = 19;
    private static final int    INSERT        = 20;
    private static final int    DELETE        = 21;

    private static final String[][] fKeys = {
        /*LINUX      VT102      ANSI        XTERM      XTERM-COLOR        */
        { "\033[1~", "\001",    "\033[H",   "\033OH",  "\033[7~" /* HOME  */},
        { "\033[4~", "\005",    "\005",     "\033OF",  "\033[8~" /* END   */},
        { "\033[5~", "\033v",   "\033v",    "\033[5~", "\033[5~" /* PGUP  */},
        { "\033[6~", "\026",    "\026",     "\033[6~", "\033[6~" /* PGDN  */},
        { "\033[A",  "\033[A",  "\033[A",   "\033[A",  "\033[A"  /* UP    */},
        { "\033[B",  "\033[B",  "\033[B",   "\033[B",  "\033[B"  /* DOWN  */},
        { "\033[D",  "\033[D",  "\033[D",   "\033[D",  "\033[D"  /* LEFT  */},
        { "\033[C",  "\033[C",  "\033[C",   "\033[C",  "\033[C"  /* RIGHT */},
        { "\033[[A", "\033OP",  null,       "\033OP",  "\033[11~"/* F1    */},
        { "\033[[B", "\033OQ",  null,       "\033OQ",  "\033[12~"/* F2    */},
        { "\033[[C", "\033OR",  null,       "\033OR",  "\033[13~"/* F3    */},
        { "\033[[D", "\033OS",  null,       "\033OS",  "\033[14~"/* F4    */},
        { "\033[[E", "\033Ot",  null,       "\033[15~","\033[15~"/* F5    */},
        { "\033[17~","\033Ou",  null,       "\033[17~","\033[17~"/* F6    */},
        { "\033[18~","\033Ov",  null,       "\033[18~","\033[18~"/* F7    */},
        { "\033[19~","\033Ol",  null,       "\033[19~","\033[19~"/* F8    */},
        { "\033[20~","\033Ow",  null,       "\033[20~","\033[20~"/* F9    */},
        { "\033[21~","\033Ox",  null,       "\033[21~","\033[21~"/* F10   */},
        { "\033[23~","\033Oy",  null,       "\033[23~","\033[23~"/* F11   */},
        { "\033[24~","\033Oz",  null,       "\033[24~","\033[24~"/* F12   */},
        { "\033[2~", null,      "\033[L",   "\033[2~", "\033[2~" /* INSERT*/},
        { "\033[3~", "\004",    "\004",     "\033[3~", "\004"    /* DELETE*/}};

    private static final boolean[] ctrlAction = {
        true, false,false,false,false,false,false,true,
        true, true, true, true, true, true, true, true,
        false,false,false,false,false,false,false,false,
        true, false,true, true, false,false,false,false };

    private static final boolean[] ctrlAlways = {
        true, false,false,false,false,false,false,false,
        true, false,true, false,true, true, true, true,
        false,false,false,false,false,false,false,false,
        false,false,false,true, false,false,false,false };

    private int       npar;
    private int[]     par                     = new int[NPAR];
    private int       width                   = -1;
    private int       height                  = -1;
    private int       cursorKeyType           = LINUX;
    private boolean   backspaceSendsDelete    = true;
    private int       top;
    private int       bottom;
    private Console   console;
    private int       isEsc;
    private boolean   needWrap;
    private boolean   autoWrapMode;
    private boolean   dispCtrl;
    private boolean   toggleMeta;
    private boolean   insertMode;
    private boolean   applKeyMode;
    private boolean   cursorKeyMode;
    private boolean   keyboardRepeatMode;
    private boolean   crLfMode;
    private boolean   metaMode;
    private boolean   offsetMode;
    private boolean   isInverted;
    private int       mouseReporting;
    private boolean   utfEnabled;
    private int       utfCount;
    private int       utfChar;
    private boolean   useG0Map;
    private char[]    G0Map;
    private char[]    G1Map;
    private char[]    translate;
    private boolean[] tabStop;
    private boolean   cursorVisibility;
    private int       cursorX;
    private int       cursorY;
    private char      attr;
    private boolean   isQuestionMark;
    private int       currentScreen;
    private int[]     savedX;
    private int[]     savedY;
    private char[]    savedAttr;
    private boolean   savedUseG0Map;
    private char[]    savedG0Map;
    private char[]    savedG1Map;
    private boolean[] savedValid;
    private String    respondString;
    private String    statusString;
    private char      lastCharacter;


    public VT102(Console console) {
        this.console = console;
        console.registerEmulation(this);
        return;
    }


    public void setSize(Dimension d) {
        if (width  == d.width &&
            height == d.height)
            return;
        width      = d.width;
        height     = d.height;
        savedX     = new int[2];
        savedY     = new int[2];
        savedAttr  = new char[2];
        savedValid = new boolean[2];
        reset();
        return;
    }


    public void setCursorKeyType(int cursorKeyType) {
        this.cursorKeyType = cursorKeyType;
        return;
    }


    public void setBackspaceType(boolean backspaceSendsDelete) {
        this.backspaceSendsDelete = backspaceSendsDelete;
        return;
    }


    public void reset() {
        isEsc              = ESnormal;
        needWrap           = false;
        autoWrapMode       = true;
        dispCtrl           = false;
        toggleMeta         = false;
        insertMode         = false;
        applKeyMode        = false;
        cursorKeyMode      = false;
        keyboardRepeatMode = true;
        crLfMode           = false;
        metaMode           = true;
        offsetMode         = false;
        isInverted         = false;
        mouseReporting     = 0;
        utfEnabled         = false;
        attr               = 0xF0;
        useG0Map           = true;
        G0Map              = Latin1Map;
        G1Map              = VT100GraphicsMap;
        translate          = G0Map;
        top                = 0;
        bottom             = height;
        cursorVisibility   = true;
        cursorX            = 0;
        cursorY            = 0;
        lastCharacter      = ' ';
        tabStop            = new boolean[width];
        for (int i = 8; i < width; i += 8)
            tabStop[i]     = true;
        currentScreen      = 0;
        console.invertScreen(false);
        console.setUTFEnabled(false);
        console.enableAlternateScreen(false);
        console.clearRegion(0,0,width,height,attr);
        console.setCursorPosition(new Point(0,0));
        console.setCursorAttr(0,0);
        console.setCursorVisibility(true);
        console.flushOperations();
        return;
    }


    private final void saveCursor() {
        savedX[currentScreen]     = cursorX;
        savedY[currentScreen]     = cursorY;
        savedAttr[currentScreen]  = attr;
        savedUseG0Map             = useG0Map;
        savedG0Map                = G0Map;
        savedG1Map                = G1Map;
        savedValid[currentScreen] = true;
        return;
    }


    private final void restoreCursor() {
        if (!savedValid[currentScreen])
            return;
        cursorX       = savedX[currentScreen];
        cursorY       = savedY[currentScreen];
        attr          = savedAttr[currentScreen];
        useG0Map      = savedUseG0Map;
        G0Map         = savedG0Map;
        G1Map         = savedG1Map;
        translate     = useG0Map ? G0Map : G1Map;
        needWrap      = false;
        return;
    }


    public String parseKeyCode(int key,int modifiers) {
        String rc      = null;
        if ((modifiers & (Event.META_MASK|Event.ALT_MASK)) != 0 &&
            (modifiers &~(Event.META_MASK|Event.ALT_MASK)) == 0) {
            // Many programs have difficulties dealing with parametrized
            // escape sequences for function keys. Thus, if ALT is the only
            // modifier key, return Emacs-style keycodes for commonly used
            // keys.
            switch (key) {
            case Event.PGUP:        rc = "\033<";                        break;
            case Event.PGDN:        rc = "\033>";                        break;
            case Event.UP:          rc = "\033p";                        break;
            case Event.DOWN:        rc = "\033n";                        break;
            case Event.LEFT:        rc = "\033b";                        break;
            case Event.RIGHT:       rc = "\033f";                        break;
            case Event.BACK_SPACE:  rc = "\027";                         break;
            case Event.DELETE:      rc = "\033d";                        break;
            default:                                                     break;
            }
        }
        if (rc == null) {           
            switch (key) {
            case Event.HOME:         rc = fKeys[HOME][cursorKeyType];    break;
            case Event.END:          rc = fKeys[END][cursorKeyType];     break;
            case Event.PGUP:         if ((modifiers & Event.SHIFT_MASK) ==
                                         Event.SHIFT_MASK)
                                       console.scrollBack();             else
                                     rc = fKeys[PGUP][cursorKeyType];    break;
            case Event.PGDN:         if ((modifiers & Event.SHIFT_MASK) ==
                                         Event.SHIFT_MASK)
                                       console.scrollFore();             else
                                     rc = fKeys[PGDN][cursorKeyType];    break;
            case Event.UP:           if (cursorKeyMode) rc = "\033OA";   else
                                     rc = fKeys[UP][cursorKeyType];      break;
            case Event.DOWN:         if (cursorKeyMode) rc = "\033OB";   else
                                     rc = fKeys[DOWN][cursorKeyType];    break;
            case Event.LEFT:         if (cursorKeyMode) rc = "\033OD";   else
                                     rc = fKeys[LEFT][cursorKeyType];    break;
            case Event.RIGHT:        if (cursorKeyMode) rc = "\033OC";   else
                                     rc = fKeys[RIGHT][cursorKeyType];   break;
            case Event.F1:           rc = fKeys[F1][cursorKeyType];      break;
            case Event.F2:           rc = fKeys[F2][cursorKeyType];      break;
            case Event.F3:           rc = fKeys[F3][cursorKeyType];      break;
            case Event.F4:           rc = fKeys[F4][cursorKeyType];      break;
            case Event.F5:           rc = fKeys[F5][cursorKeyType];      break;
            case Event.F6:           rc = fKeys[F6][cursorKeyType];      break;
            case Event.F7:           rc = fKeys[F7][cursorKeyType];      break;
            case Event.F8:           rc = fKeys[F8][cursorKeyType];      break;
            case Event.F9:           rc = fKeys[F9][cursorKeyType];      break;
            case Event.F10:          rc = fKeys[F10][cursorKeyType];     break;
            case Event.F11:          rc = fKeys[F11][cursorKeyType];     break;
            case Event.F12:          rc = fKeys[F12][cursorKeyType];     break;
            case Event.INSERT:       rc = fKeys[INSERT][cursorKeyType];  break;
            case Event.DELETE:       rc = fKeys[DELETE][cursorKeyType];  break;
            case Event.ENTER:        rc = applKeyMode? "\033OM" :
                                          crLfMode ? "\r\n" : "\n";      break;
            case Event.BACK_SPACE:   rc = backspaceSendsDelete ?
                                          "\177" : "\010";               break;
            case Event.TAB:          rc = ((modifiers & Event.SHIFT_MASK)!=0) ?
                                          "\033[Z" : "\011";
                                     modifiers &= ~Event.SHIFT_MASK;     break;
            case Event.ESCAPE:       rc = "\033";                        break;
            default:                                                     break;
            }
        }
        // Some of the escape sequences can take parameters that specify the
        // modifier keys. Be careful with this, many termcap/terminfo
        // databases neglect to mention this feature!
        if (rc != null &&
            (modifiers & (Event.SHIFT_MASK|Event.CTRL_MASK|Event.META_MASK|
                          Event.ALT_MASK)) != 0) {
            String part1 = null;
            String part2 = null;
            if (rc.startsWith("\033[[")) {
                part1 = "\033[[";
                part2 = rc.substring(3);
            } else if (rc.startsWith("\033[")) {
                for (part1 = "\033[";
                     part1.length() < rc.length() &&
                     Character.isDigit(rc.charAt(part1.length())); )
                    part1 = rc.substring(0,part1.length() + 1);
                part2 = rc.substring(part1.length());
                if (part1.length() > 2)
                    part1 += ";";
            } else if (rc.startsWith("\033O")) {
                part1 = "\033O";
                part2 = rc.substring(2);
            }
            if (part1 != null) {
                // SHIFT sets bit 0
                // ALT   sets bit 1
                // CTRL  sets bit 2
                // if any modifier key is pressed, the total value is the
                // sum of the bits plus one
                int mod =((modifiers & Event.SHIFT_MASK) != 0 ? 1 : 0) +
                         ((modifiers&(Event.ALT_MASK|Event.META_MASK))!=0?2:0)+
                         ((modifiers & Event.CTRL_MASK) != 0 ?  4 : 0) +
                         1;
                rc = part1 + mod + part2;
            }
        }
        return(rc);
    }


    public boolean getMetaMode() {
        return(metaMode);
    }


    public String writeString(String s) {
        respondString = "";
        StringBuffer lineBuf = new StringBuffer();
        int inputLength      = s.length();
        console.setCursorVisibility(false);
        for (int i = 0; i < inputLength; i++) {
            char ch = s.charAt(i);
            boolean isNormalCharacter =
                (ch >= ' ' ||
                 (!utfEnabled && !(dispCtrl ? ctrlAlways : ctrlAction)[ch])) &&
                (ch != '\u007F' || dispCtrl) &&
                (utfEnabled || ch != '\u009B');
            if (utfEnabled) {
                if (ch > '\u007F') {
                    if (utfCount > 0 && (ch & 0xc0) == 0x80) {
                        utfChar = (utfChar << 6) | (ch & 0x3f);
                        if (--utfCount <= 0) {
                            if (utfChar > 0xFFFF || utfChar < 0)
                                ch = UNKNOWN_GLYPH;
                            else
                                ch = (char)utfChar;
                        } else
                            continue;
                    } else {
                        if ((ch & 0xe0) == 0xc0) {
                            utfCount = 1; utfChar = ch & 0x1f;
                        } else if ((ch & 0xf0) == 0xe0) {
                            utfCount = 2;
                            utfChar  = ch & 0x0f;
                        } else if ((ch & 0xf8) == 0xf0) {
                            utfCount = 3; utfChar = ch & 0x07;
                        } else if ((ch & 0xfc) == 0xf8) {
                            utfCount = 4; utfChar = ch & 0x03;
                        } else if ((ch & 0xfe) == 0xfc) {
                            utfCount = 5; utfChar = ch & 0x01;
                        } else
                            utfCount = 0;
                        continue;
                    }
                } else
                    utfCount = 0;
            }
            if (isNormalCharacter && isEsc == ESnormal) {
                if (!utfEnabled && ch < '\u0100')
                    ch = translate[toggleMeta ? (ch | '\u0080') : ch];
                if ((ch & 0xFF00) == 0xF000)
                    ch = (char)(ch & 0xFF);
                else if (ch == '\uFEFF' ||
                         (ch >= '\u200A' && ch <= '\u200F'))
                         continue;
                if (needWrap || insertMode)
                    renderString(lineBuf);
                if (needWrap) {
                    cr(); lf(1);
                }
                if (insertMode)
                    console.scrollRegion(cursorX,cursorY,width-cursorX-1,1,
                                         1,0,attr);
                lineBuf.append(ch);
                lastCharacter = ch;
                if (cursorX+lineBuf.length() >= width) {
                    needWrap = autoWrapMode;
                }
            } else {
                renderString(lineBuf);
                doControl(ch,lineBuf);
            }
        }
        renderString(lineBuf);
        console.setCursorPosition(new Point(cursorX,cursorY));
        console.setCursorVisibility(cursorVisibility);
        console.flushOperations();
        return(respondString);
    }


    private final void renderString(StringBuffer lineBuf) {
        int incX = lineBuf.length();
        if (incX > 0) {
            if (incX > width-cursorX) {
                char lastCharacter = lineBuf.charAt(incX-1);
                incX = width-cursorX;
                lineBuf.setLength(incX);
                lineBuf.setCharAt(incX-1,lastCharacter);
            }
            console.drawString(lineBuf.toString(),
                               cursorX,cursorY,attr);
            cursorX += incX;
            if (cursorX >= width) {
                cursorX = width-1;
            }
        }
        lineBuf.setLength(0);
        return;
    }


    private final void doControl(char ch,StringBuffer lineBuf) {
        switch (ch) {
        case '\u0000':                                return;
        case '\u0007': if (isEsc == ESstatus)         break;
                       console.beep();                return;
        case '\u0008': bs();                          return;
        case '\u0009': while (cursorX < width-1 && !tabStop[++cursorX]);
                                                      return;
        case '\n':
        case '\u000B':
        case '\u000C': lf(1); if (!crLfMode)          return;
        case '\r':     cr();                          return;
        case '\u000E': useG0Map  = false;
                       translate = G1Map;
                       dispCtrl  = true;              return;
        case '\u000F': useG0Map  = true;
                       translate = G0Map;
                       dispCtrl  = false;             return;
        case '\u0018':
        case '\u001A': isEsc     = ESnormal;          return;
        case '\u001B': isEsc     = ESesc;             return;
        case '\u007F': del();                         return;
        case '\u009B': isEsc     = ESsquare;          return;
        default:                                      break;
        }
        switch (isEsc) {
        case ESesc:
            isEsc = ESnormal;
            switch (ch) {
            case '[': isEsc = ESsquare;               break;
            case ']': isEsc = ESnonstd;               break;
            case '%': isEsc = ESpercent;              break;
            case 'E': cr(); lf(1);                    break;
            case 'M': ri(1);                          break;
            case 'D': lf(1);                          break;
            case 'H': tabStop[cursorX] = true;        break;
            case 'Z': respondID();                    break;
            case '7': saveCursor();                   break;
            case '8': restoreCursor();                break;
            case '(': isEsc = ESsetG0;                break;
            case ')': isEsc = ESsetG1;                break;
            case '#': isEsc = EShash;                 break;
            case 'g': console.invertScreen(!isInverted);
                      console.invertScreen( isInverted);
                                                      break;
            case 'c': reset();                        break;
            case '>': applKeyMode = false;            break;
            case '=': applKeyMode = true;             break;
            default:                                  break;
            }
            break;
        case ESnonstd:
            switch (ch) {
            case 'P': for (int i = NPAR; i-- > 0; ) par[i] = 0;
                      npar   = 0; isEsc  = ESpalette; break;
            case 'R': // palette support is not implemented! /***/
                      isEsc  = ESnormal;              break;
            case '1':
            case '2': statusString = "";
                      isEsc  = ESstatus;              break;
            default:  isEsc  = ESnormal;              break;
            }
            break;
        case ESpalette:
            if ((ch >= '0' && ch <= '9') ||
                (ch >= 'A' && ch <= 'F') || (ch >= 'a' && ch <= 'f')) {
                par[npar++] = (ch > '9')?((ch&'\u00DF') - 'A' + '\n'):(ch-'0');
                if (npar == 7) {
                    // palette support is not implemented! /***/
                    isEsc = ESnormal;
                }
            } else
                isEsc = ESnormal;
            break;
        case ESsquare:
            for (int i = NPAR; i-- > 0; )
                par[i] = 0;
            npar   = 0;
            isEsc  = ESgetpars;
            if (ch == '[') { // function key
                isEsc = ESfunckey;
                return;
            }
            isQuestionMark = (ch == '?');
            if (isQuestionMark)
                return;
            // fall thru
        case ESgetpars:
            if (ch == ';' && npar < NPAR-1) {
                npar++;
                return;
            } else if (ch >= '0' && ch <= '9') {
                par[npar] = 10*par[npar] + (ch - '0');
                return;
            } else
                isEsc = ESgotpars;
            // fall thru
        case ESgotpars:
            isEsc = ESnormal;
            switch (ch) {
            case 'h': setMode(true);                  return;
            case 'l': setMode(false);                 return;
            case 'c':
                if (isQuestionMark) {
                    console.setCursorAttr(par[2],par[1]);
                    return;
                }
                break;
            case 'm':
                if (isQuestionMark) {
                    return;
                }
                break;
            case 'n':
                if (!isQuestionMark) {
                    switch (par[0]) {
                    case 5: statusReport();           break;
                    case 6: cursorReport();           break;
                    default:                          break;
                    }
                }
                return;
            default:                                  break;
            }
            if (isQuestionMark) {
                isQuestionMark = false;
                return;
            }
            switch (ch) {
            case 'G':
            case '`': if (par[0] != 0) par[0]--;
                      gotoXY(par[0],cursorY);         break;
            case 'A': if (par[0] == 0) par[0]++;
                      gotoXY(cursorX,cursorY-par[0]); break;
            case 'B':
            case 'e': if (par[0] == 0) par[0]++;
                      gotoXY(cursorX,cursorY+par[0]); break;
            case 'C':
            case 'a': if (par[0] == 0) par[0]++;
                      gotoXY(cursorX+par[0],cursorY); break;
            case 'D': if (par[0] == 0) par[0]++;
                      gotoXY(cursorX-par[0],cursorY); break;
            case 'E': if (par[0] == 0) par[0]++;
                      gotoXY(0,cursorY+par[0]);       break;
            case 'F': if (par[0] == 0) par[0]++;
                      gotoXY(0,cursorY-par[0]);       break;
            case 'd': if (par[0] != 0) par[0]--;
                      gotoXaY(cursorX,par[0]);        break;
            case 'H':
            case 'f': if (par[0] != 0) par[0]--;
                      if (par[1] != 0) par[1]--;
                      gotoXaY(par[1],par[0]);         break;
            case 'I': if (par[0] == 0) par[0]++;
                      while (cursorX < width-1 &&
                             (!tabStop[++cursorX] || --par[0] > 0));
                                                      break;
            case 'J': csiJ(par[0]);                   break;
            case 'K': csiK(par[0]);                   break;
            case 'L': csiL(par[0]);                   break;
            case 'M': csiM(par[0]);                   break;
            case 'P': csiP(par[0]);                   break;
            case 'S': if (par[0] == 0) par[0]++;
                      lf(par[0]);                     break;
            case 'T': if (par[0] == 0) par[0]++;
                      ri(par[0]);                     break;
            case 'c': if (par[0] == 0) respondID();   break;
            case 'g': if (par[0] == 0) tabStop[cursorX] = false;
                      else if (par[0] == 2 || par[0] == 3)
                          tabStop = new boolean[width];
                                                      break;
            case 'm': csim();                         break;
            case 'q': // LED control not implemented! /***/
                                                      break;
            case 'r': if (par[0] == 0) par[0]++;
                      if (par[1] == 0) par[1] = height;
                      if (par[0] < par[1] && par[1] <= height) {
                          top    = par[0] - 1;
                          bottom = par[1];
                          gotoXaY(0,0);
                      }                               break;
            case 'b': if (par[0] == 0) par[0]++;
                      else if (par[0] > width*height)
                          par[0] = width*height;
                      while (par[0]-- > 0) lineBuf.append(lastCharacter);
                                                      break;
            case 's': saveCursor();                   break;
            case 'u': restoreCursor();                break;
            case 'X': csiX(par[0]);                   break;
            case '@': csiAt(par[0]);                  break;
            case 'Z': while (cursorX > 0 && (!tabStop[--cursorX]
                             || --par[0] > 0));       break;
            case ']': settermCommand();               break;
            default:                                  break;
            }
            break;
        case ESpercent:
            isEsc = ESnormal;
            switch (ch) {
            case '@': utfEnabled = false;             break;
            case 'G':
            case '8': utfEnabled = true;              break;
            default:                                  break;
            }
            console.setUTFEnabled(utfEnabled);
            break;
        case ESfunckey: isEsc = ESnormal;             break;
        case EShash:
            isEsc = ESnormal;
            if (ch == '8') {
                // screen alignment test not implemented! /***/
            }
            break;
        case ESsetG0:
            switch (ch) {
            case '0': G0Map = VT100GraphicsMap;       break;
            case 'B': G0Map = Latin1Map;              break;
            case 'U': G0Map = CodePage437Map;         break;
            case 'K': G0Map = DirectToFontMap;        break;
            default:                                  break;
            }
            if (useG0Map)
                translate = G0Map;
            isEsc = ESnormal;
            break;
        case ESsetG1:
            switch (ch) {
            case '0': G1Map = VT100GraphicsMap;       break;
            case 'B': G1Map = Latin1Map;              break;
            case 'U': G1Map = CodePage437Map;         break;
            case 'K': G1Map = DirectToFontMap;        break;
            default:                                  break;
            }
            if (!useG0Map)
                translate = G1Map;
            isEsc = ESnormal;
            break;
        case ESstatus:
            if (ch == '\007') {
                if (statusString.startsWith(";"))
                    statusString = statusString.substring(1);
                console.displayStatus(statusString);
                isEsc = ESnormal;
            } else
                statusString += ch;
            break;
        default: isEsc = ESnormal;                    break;
        }
        return;
    }


    private final void cr() {
        cursorX  = 0;
        needWrap = false;
        return;
    }


    private final void lf(int number) {
        if (number > height)
            number = height;
        while (number-- > 0) {
            if (cursorY == bottom-1)
                console.scrollRegion(0,top+1,width,bottom-top-1,0,-1,attr);
            else if (cursorY < height-1)
                cursorY++;
        }
	needWrap = false;
        return;
    }


    private final void bs() {
        if (cursorX > 0) {
            cursorX--;
            needWrap = false;
        }
        return;
    }


    private final void del() {
        // ignored
        return;
    }


    private final void ri(int number) {
        if (number > height)
            number = height;
        while (number-- > 0) {
            if (cursorY == top)
                console.scrollRegion(0,top,width,bottom-top-1,0,1,attr);
            else if (cursorY > 0)
		cursorY--;
        }
	needWrap = false;
        return;
    }


    private final void gotoXY(int x,int y) {
        if (x < 0)
            cursorX = 0;
        else if (x >= width)
            cursorX = width-1;
        else
            cursorX = x;
        int minY, maxY;
        if (offsetMode) {
            minY = top;
            maxY = bottom;
        } else {
            minY = 0;
            maxY = height;
        }
        if (y < minY)
            cursorY = minY;
        else if (y >= maxY)
            cursorY = maxY-1;
        else
            cursorY = y;
        needWrap = false;
        return;
    }


    private final void gotoXaY(int x,int y) {
        gotoXY(x,offsetMode ? (top + y) : y);
        return;
    }


    private final void respondID() {
        respondString += VT102ID;
        return;
    }


    private final void statusReport() {
        respondString += "\u001B[0n"; // terminal OK
        return;
    }


    private final void cursorReport() {
        respondString += "\u001B[" +
                         (cursorY + (offsetMode ? top + 1 : 1)) +
                         ";" +
                         (cursorX + 1) +
                         "R";
        return;
    }


    private final void setMode(boolean state) {
        for (int i = 0; i <= npar; i++)
            if (isQuestionMark)
                switch (par[i]) {
                case 1:
                    cursorKeyMode = state;
                    break;
                case 3:
                    // toggling between 80/132 mode is not implemented! /***/
                    break;
                case 5:
                    if (isInverted != state) {
                        isInverted = state;
                        console.invertScreen(state);
                    }
                    break;
                case 6:
                    offsetMode = state;
                    gotoXaY(0,0);
                    break;
                case 7:
                    autoWrapMode = state;
                    break;
                case 8:
                    keyboardRepeatMode = state;
                    break;
                case 9:
                    mouseReporting = (state ? 1 : 0);
                    break;
                case 25:
                    cursorVisibility = state;
                    break;
                case 47:
                    currentScreen = (state ? 1 : 0);
                    console.enableAlternateScreen(state);
                    break;
                case 1000:
                    mouseReporting = (state ? 2 : 0);
                    break;
                default:
                    break;
                }
            else switch (par[i]) {
            case 3:
                dispCtrl = state;
                break;
            case 4:
                insertMode = state;
                break;
            case 20:
                crLfMode = state;
                break;
            default:
                break;
            }
        return;
    }


    private final void settermCommand() {
        // setterm commands are not implemented! /***/
        return;
    }


    private final void csiJ(int number) {
        switch (number) {
        case 0: // erase from cursor to end of display
            console.clearRegion(cursorX,cursorY,
                                width-cursorX,1,attr);
            if (cursorY < height-2)
                console.clearRegion(0,cursorY+1,
                                    width,height-cursorY-1,attr);
            break;
        case 1: // erase from start to cursor
            if (cursorY > 0)
                console.clearRegion(0,0,width,cursorY,attr);
            console.clearRegion(0,cursorY,cursorX+1,1,attr);
            break;
        case 2: // erase whole display
            console.clearRegion(0,0,width,height,attr);
            break;
        default:
            return;
        }
        needWrap = false;
        return;
    }


    private final void csiK(int number) {
        switch (number) {
        case 0: // erase from cursor to end of line
            console.clearRegion(cursorX,cursorY,width-cursorX,1,attr);
            break;
        case 1: // erase from start of line to cursor
            console.clearRegion(0,cursorY,cursorX+1,1,attr);
            break;
        case 2: // erase whole line
            console.clearRegion(0,cursorY,width,1,attr);
            break;
        default:
            return;
        }
        needWrap = false;
        return;
    }


    private final void csiL(int number) {
        if (cursorY >= bottom)
            return;
        if (number == 0)
            number = 1;
        if (number > bottom - cursorY)
            number = bottom - cursorY;
        console.scrollRegion(0,cursorY,width,bottom-cursorY-number,
                             0,number,attr);
        needWrap = false;
        return;
    }


    private final void csiM(int number) {
        if (cursorY >= bottom)
            return;
        if (number == 0)
            number = 1;
        if (number > bottom - cursorY)
            number = bottom - cursorY;
        console.scrollRegion(0,cursorY+number,width,bottom-cursorY-number,
                             0,-number,attr);
        needWrap = false;
        return;
    }


    private final void csiP(int number) {
        if (number == 0)
            number = 1;
        if (number > width - cursorX)
            number = width - cursorX;
        console.scrollRegion(cursorX+number,cursorY,width-cursorX-number,1,
                             -number,0,attr);
        needWrap = false;
        return;
    }


    private final void csim() {
        for (int i = 0; i <= npar; i++)
            switch (par[i]) {
            case 0:
                attr = 0xF0;
                break;
            case 1:
                attr = (char)((attr & ~Console.DIM)|
                              Console.BRIGHT);
                break;
            case 2:
                attr = (char)((attr & ~Console.BRIGHT)|
                              Console.DIM);
                break;
            case 4:
                attr |= Console.UNDERLINE;
                break;
            case 5:
                attr |= Console.BLINK;
                break;
            case 7:
                attr |= Console.REVERSE;
                break;
            case 10:
                translate  = useG0Map ? G0Map : G1Map;
                dispCtrl   = false;
                toggleMeta = false;
                break;
            case 11:
                translate  = CodePage437Map;
                dispCtrl   = true;
                toggleMeta = false;
                break;
            case 12:
                translate  = CodePage437Map;
                dispCtrl   = true;
                toggleMeta = true;
                return;
            case 21:
            case 22:
                attr &=~(Console.BRIGHT|Console.DIM);
                break;
            case 24:
                attr &=~Console.UNDERLINE;
                break;
            case 25:
                attr &=~Console.BLINK;
                break;
            case 27:
                attr &=~Console.REVERSE;
                break;
            case 38:
                attr = (char)((attr &~(Console.DIM|
                                       Console.BRIGHT|
                                       0x0F))|
                              Console.UNDERLINE);
                                    
            case 39:
                attr = (char)(attr &
                              ~(Console.DIM|
                                Console.BRIGHT|
                                Console.UNDERLINE|
                                0x0F));
                break;
            case 49:
                attr = (char)(attr | 0xF0);
                break;
            default:
                if (par[i] >= 30 && par[i] <= 37) {
                    int fg = par[i] - 30;
                    attr = (char)((attr &~0x0F) | fg);
                } else if (par[i] >= 40 && par[i] <= 47) {
                    int bg = par[i] - 40;
                    attr = (char)((attr &~0xF0) | (bg<<4));
                }
                break;
            }
        return;
    }


    private final void csiX(int number) {
        if (number == 0)
            number++;
        if (number > width - cursorX)
            number = width - cursorX;
        console.clearRegion(cursorX,cursorY,number,1,attr);
        needWrap = false;
        return;
    }


    private final void csiAt(int number) {
        if (number == 0)
            number = 1;
        if (number > width - cursorX)
            number = width - cursorX;
        console.scrollRegion(cursorX,cursorY,width-cursorX-number,1,
                             number,0,attr);
        needWrap = false;
        return;
    }


    public String processMouseEvent(int event,int x,int y,int clickCount,
                                    int button,int modifiers) {
        if (mouseReporting == 0 || (modifiers & Event.SHIFT_MASK) != 0)
            return(null);
        switch (event) {
        case Event.MOUSE_UP:
        case Event.MOUSE_DOWN:
            button = (event == Event.MOUSE_UP) ? 3 :
                     (button == 1)             ? 0 :
                     (button == 2)             ? 1 : 2;
            if (event == Event.MOUSE_DOWN &&
                ((modifiers & (Event.ALT_MASK | Event.META_MASK)) != 0))
                button += 8;
            if (event == Event.MOUSE_DOWN &&
                ((modifiers & (Event.CTRL_MASK)) != 0))
                button += 16;
            return("\033[M" + (char)(button + ' ') +
                              (char)(x      + '!') +
                              (char)(y      + '!'));
        case Event.MOUSE_MOVE:
        case Event.MOUSE_DRAG:
        case Event.MOUSE_ENTER:
        case Event.MOUSE_EXIT:
        default:
            return("");
        }
    }
}

/*
 * Local variables:
 * mode: java
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
