#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <syslog.h>
#include <time.h>

#include "connectionmgr.h"
#include "shellinaboxd.h"


struct Connection {
    int        fd;
    int        oldEvents;
    int        pollIdx;
    int        deleted;
    int        hasTimeOut;
    time_t     timeOut;
    ServerData *serverData;
};


static Connection    **connections  = NULL;
static int           numConnections = 0;
static struct pollfd *pConnections  = NULL;
static time_t        curTime        = 0;
static time_t        tmo            = -1;


void initConnections(void)
{
    time(&curTime);
    return;
}


Connection *addConnection(int fd,ServerData *serverData)
{
    Connection    *newConnection;
    struct pollfd *newPollConnection;
    int           idx = -1,min,max;

    // Modified binary search, to locate insertion position
    if (numConnections == 0)
        idx = 0;
    else {
        min = 0;
        max = numConnections-1;
        if (fd > connections[max]->fd)
            idx = numConnections;
        else if (fd < connections[0]->fd)
            idx = 0;
        else {
            while (min < max) {
                idx = (min + max)/2;
                if (fd < connections[idx]->fd)
                    max = idx-1;
                else if (fd == connections[idx]->fd)
                    return(NULL);
                else
                    min = ++idx; } } }
    // Straight-insertion really happens to be a good choice for a sorting
    // algorithm, because adding/removing connections happens only
    // very infrequently and because we need arrays to quickly map between
    // the results from poll and the associated connection
    connections = realloc(connections,sizeof(Connection *) * ++numConnections);
    memmove(connections + idx + 1,connections + idx,
            (numConnections - idx - 1)*sizeof(Connection *));
    pConnections= realloc(pConnections,sizeof(struct pollfd) * numConnections);
    memmove(pConnections + idx + 1,pConnections + idx,
            (numConnections - idx - 1)*sizeof(struct pollfd));
    // Create new entries
    newConnection = 
        connections[idx]      = malloc(sizeof(Connection));
    newConnection->fd         = fd;
    newConnection->oldEvents  = 0;
    newConnection->pollIdx    = idx;
    newConnection->deleted    = 0;
    newConnection->serverData = serverData;
    newConnection->hasTimeOut = 0;
    newPollConnection         = pConnections + idx;
    newPollConnection->fd     = fd;
    newPollConnection->events = 0;
    return(newConnection);
}


Connection *findConnection(int fd,ServerData **serverData)
{
    int idx,min,max;

    // Binary search
    min = 0;
    max = numConnections-1;
    while (min <= max) {
        idx = (min + max)/2;
        if (fd < connections[idx]->fd)
            max = idx-1;
        else if (fd == connections[idx]->fd) {
            if (serverData != NULL)
                *serverData = connections[idx]->serverData;
            return(connections[idx]); }
        else
            min = idx+1; }
    return(NULL);
}


void markConnectionForDeletion(int fd)
{
    Connection *connection = findConnection(fd,NULL);

#ifdef DEBUGCONNECTIONS
    syslog(LOG_DEBUG|LOG_DAEMON,"Mark for deletion: fd=%d, idx=%d",
           fd,connection != NULL ? connection->pollIdx : -1);
#endif
    if (connection != NULL && !connection->deleted) {
        connection->deleted = 1; }
    return;
}


int deleteMarkedConnections(void (*destroyFnc)(int fd,Connection *connection,
                                               ServerData *serverData))
{
    int        done;
    int        i;
    Connection *connection;

    do {
        done = 1;
        tmo = -1;
        for (i = numConnections - 1; i >= 0;) {
            connection = connections[i];
            if (connection->deleted) {
                if (destroyFnc != NULL)
                    destroyFnc(connections[i]->fd,connection,
                               connection->serverData);
                // Free data structures
                numConnections--;
                memmove(connections+i,connections+i+1,
                        (numConnections-i)*sizeof(Connection *));
                memmove(pConnections+i,pConnections+i+1,
                        (numConnections-i)*sizeof(struct pollfd));
                free(connection);
                if (i >= numConnections)
                    i--; }
            else {
                // Compute next timeout event
                if (connection->hasTimeOut) {
                    int seconds = connection->timeOut - curTime;
                    if (seconds < 0)
                        seconds = 0;
                    if (tmo < 0 || seconds < tmo)
                        tmo = seconds; }
                i--; } }
        for (i = numConnections; i-- > 0;) {
            int newEvents = pConnections[i].events&~connections[i]->oldEvents;
            connections[i]->oldEvents = pConnections[i].events;
            connections[i]->pollIdx = i;
            // All our I/O is non-blocking, thus it pays off to try operations
            // even before calling poll
            if (newEvents & POLLIN) {
                FNC(connections[i]->serverData,PollIn)(connections[i]->fd,
                                connections[i],&connections[i]->serverData->u);
                done = 0; }
            if (newEvents & POLLOUT) {
                FNC(connections[i]->serverData,PollOut)(connections[i]->fd,
                                connections[i],&connections[i]->serverData->u);
                done = 0; } }
    } while (!done);
    return(numConnections);
}


Connection *firstConnection(int *fd,ServerData **serverData)
{
    Connection *connection;

    if (numConnections <= 0)
        return(NULL);
    connection = connections[0];
    if (fd != NULL)
        *fd = connection->fd;
    if (serverData != NULL)
        *serverData = connection->serverData;
    return(connection);
}


Connection *nextConnection(Connection *connection,
                           int *fd,ServerData **serverData)
{
    if (connection->pollIdx+1 >= numConnections)
        return(NULL);
    connection = connections[connection->pollIdx+1];
    if (fd != NULL)
        *fd = connection->fd;
    if (serverData != NULL)
        *serverData = connection->serverData;
    return(connection);
}


void setConnectionTimeOut(Connection *connection,int timeOut)
{
    if (timeOut >= 0) {
        connection->timeOut    = curTime + timeOut;
        connection->hasTimeOut = 1; }
    else
        connection->hasTimeOut = 0;
    return;
}


int isTimedOut(Connection *connection)
{
    return(connection->timeOut - curTime <= 0);
}


void enablePollEvents(Connection *connection,int events)
{
    pConnections[connection->pollIdx].events |= events;
    return;
}


void disablePollEvents(Connection *connection,int events)
{
    pConnections[connection->pollIdx].events &= ~events;
    return;
}


int pollConnections(void)
{
    int rc = poll(pConnections,numConnections,tmo*1000);
    time(&curTime);
#ifdef DEBUGCONNECTIONS
    {extern Fnc cgiCommVTable[],serverVTable[],sessionVTable[];
    char buf[16384];
    int   i;
    sprintf(buf,"num=%d, tmo=%d, rc=%d ",
            numConnections,(int)tmo,rc);
    for (i = 0; i < numConnections; i++) {
        if (connections[i]->pollIdx != i)
            sprintf(strrchr(buf,'\000'),"(%d!=%d)",connections[i]->pollIdx,i);
        sprintf(strrchr(buf,'\000'),"[%d",connections[i]->fd);
        if (connections[i]->hasTimeOut)
            sprintf(strrchr(buf,'\000'),",%ds",
                    (int)(connections[i]->timeOut-curTime));
        if (pConnections[i].events != 0) {
            char *sep = ",";
            if (pConnections[i].events&POLLIN) {
                sprintf(strrchr(buf,'\000'),"%sPOLLIN",  sep); sep = "|"; }
            if (pConnections[i].events&POLLPRI) {
                sprintf(strrchr(buf,'\000'),"%sPOLLPRI", sep); sep = "|"; }
            if (pConnections[i].events&POLLOUT) {
                sprintf(strrchr(buf,'\000'),"%sPOLLOUT", sep); sep = "|"; }
            if (pConnections[i].events&POLLERR) {
                sprintf(strrchr(buf,'\000'),"%sPOLLERR", sep); sep = "|"; }
            if (pConnections[i].events&POLLHUP) {
                sprintf(strrchr(buf,'\000'),"%sPOLLHUP", sep); sep = "|"; }
            if (pConnections[i].events&POLLNVAL) {
                sprintf(strrchr(buf,'\000'),"%sPOLLNVAL",sep); sep = "|"; }
            if (pConnections[i].events&
                ~(POLLIN|POLLPRI|POLLOUT|POLLERR|POLLHUP|POLLNVAL))
                sprintf(strrchr(buf,'\000'),"%s%d",sep,
                        pConnections[i].events&
                        ~(POLLIN|POLLPRI|POLLOUT|POLLERR|POLLHUP|POLLNVAL)); }
        if (pConnections[i].revents != 0) {
            char *sep = " -> ";
            if (pConnections[i].revents&POLLIN) {
                sprintf(strrchr(buf,'\000'),"%sPOLLIN",  sep); sep = "|"; }
            if (pConnections[i].revents&POLLPRI) {
                sprintf(strrchr(buf,'\000'),"%sPOLLPRI", sep); sep = "|"; }
            if (pConnections[i].revents&POLLOUT) {
                sprintf(strrchr(buf,'\000'),"%sPOLLOUT", sep); sep = "|"; }
            if (pConnections[i].revents&POLLERR) {
                sprintf(strrchr(buf,'\000'),"%sPOLLERR", sep); sep = "|"; }
            if (pConnections[i].revents&POLLHUP) {
                sprintf(strrchr(buf,'\000'),"%sPOLLHUP", sep); sep = "|"; }
            if (pConnections[i].revents&POLLNVAL) {
                sprintf(strrchr(buf,'\000'),"%sPOLLNVAL",sep); }
            if (pConnections[i].revents&
                ~(POLLIN|POLLPRI|POLLOUT|POLLERR|POLLHUP|POLLNVAL))
                sprintf(strrchr(buf,'\000'),"%s%d",sep,
                        pConnections[i].revents&
                        ~(POLLIN|POLLPRI|POLLOUT|POLLERR|POLLHUP|POLLNVAL)); }
        if (connections[i]->serverData->vTable == cgiCommVTable) {
            sprintf(strrchr(buf,'\000'),",CGI");
            if (connections[i]->serverData->u.cg.done)
                sprintf(strrchr(buf,'\000'),",done!"); }
        else if (connections[i]->serverData->vTable == serverVTable) {
            sprintf(strrchr(buf,'\000'),",SERVER"); }
        else if (connections[i]->serverData->vTable == sessionVTable) {
            sprintf(strrchr(buf,'\000'),",SESSION,\"%s\",pid:%d,%d,%d",
#ifdef USELOGIN
                    connections[i]->serverData->u.ss.ptyName,
#else
                    "",
#endif
                    connections[i]->serverData->u.ss.pid,
                    connections[i]->serverData->u.ss.inputSequenceNumber,
                    connections[i]->serverData->u.ss.outputSequenceNumber); }
        else {
            sprintf(strrchr(buf,'\000'),",UNKNOWN"); }
        sprintf(strrchr(buf,'\000'),"] "); }
    syslog(LOG_DAEMON|LOG_DEBUG,"%s",buf); }
#endif
    return(rc);
}


int getServerEvents(void)
{
    return(pConnections->revents);
}


int getConnectionEvents(Connection *connection)
{
    return(pConnections[connection->pollIdx].revents);
}

/*
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
