#define _GNU_SOURCE

#include <errno.h>
#include <grp.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <syslog.h>
#include <unistd.h>
#include <utmp.h>

#include "connectionmgr.h"
#include "fifo.h"
#include "session.h"
#include "shellinaboxd.h"


static void inputAvailableCallBack(void *data)
{
    ServerData *serverData;
    Connection *connection = findConnection((int)data,&serverData);

    enablePollEvents(connection,POLLOUT);
    return;
}


static void inputEmptyCallBack(void *data)
{
    ServerData *serverData;
    Connection *connection = findConnection((int)data,&serverData);

    disablePollEvents(connection,POLLOUT);
    return;
}


static void hasSpaceCallBack(void *data)
{
    ServerData *serverData;
    Connection *connection = findConnection((int)data,&serverData);

    enablePollEvents(connection,POLLIN);
    return;
}


static void fullCallBack(void *data)
{
    ServerData *serverData;
    Connection *connection = findConnection((int)data,&serverData);

    disablePollEvents(connection,POLLIN);
    return;
}


Connection *sessionCreate(int pid,int pty,char *ptyName,int *sequenceNumber,
                          struct utmpx *utmpx)
{
    extern Fnc sessionVTable[];
    ServerData *serverData;
    Connection *connection;

    *sequenceNumber                       = random();
    serverData                            = malloc(sizeof(ServerData));
    serverData->vTable                    = sessionVTable;
    serverData->u.ss.pendingPoll          = NULL;
#ifdef USELOGIN
    strcpy(serverData->u.ss.ptyName,ptyName);
    memcpy(&serverData->u.ss.utmpx,&utmpx,sizeof(struct utmpx));
#endif
    serverData->u.ss.pid                  = pid;
    serverData->u.ss.inputSequenceNumber  = *sequenceNumber;
    serverData->u.ss.outputSequenceNumber = *sequenceNumber;
    connection                            = addConnection(pty,serverData);
    serverData->u.ss.inputFifo            = createNewFifo(0,(void *)pty,
                                                        inputAvailableCallBack,
                                                        inputEmptyCallBack,
                                                        NULL,NULL,NULL);
    serverData->u.ss.outputFifo           = createNewFifo(BLOCKSIZE/2,
                                                        (void *)pty,NULL,NULL,
                                                        hasSpaceCallBack,
                                                        fullCallBack,NULL);
    setConnectionTimeOut(connection,CONNTIMEOUT);
    return(connection);
}


static void processPendingPoll(Connection *connection,Session *session)
{
    char *chunk;
    int  count;

    // If there is a queued poll request, satisfy it now
    if (session->pendingPoll != NULL &&
        (chunk = fifoGetFirstChunk(session->outputFifo,&count)) != NULL) {
        static const char *hdr = "Content-type: "WLCONTENTTYPE"\r\n"
                                 "\r\n"DATA;
        char              *newChunk;
        newChunk = fifoCreateChunk(strlen(hdr));
        memcpy(newChunk,hdr,strlen(hdr));
        fifoAppendChunk(session->pendingPoll,newChunk,strlen(hdr));
        do {
            newChunk = fifoCreateChunk(count);
            memcpy(newChunk,chunk,count);
            fifoAppendChunk(session->pendingPoll,newChunk,count);
            fifoDequeueChunk(session->outputFifo,count);
        }while ((chunk = fifoGetFirstChunk(session->outputFifo,&count))!=NULL);
        fifoFlushAndClose(session->pendingPoll);
        destroyFifo(session->pendingPoll); /* decrease use count */
        session->pendingPoll = NULL;
        setConnectionTimeOut(connection,CONNTIMEOUT); }
    return;
}


static void sessionPollIn(int pty,Connection *connection,Session *session)
{
    char *chunk;
    int  count;

    // Our client has written data to the pty; read it into the fifo
    for (;;) {
        chunk = fifoCreateChunk(BLOCKSIZE);
        count = read(pty,chunk,BLOCKSIZE);
        if (count <= 0) {
            if (count == 0 || errno == EAGAIN) {
                fifoAppendChunk(session->outputFifo,chunk,0);
                break; }
            fifoAppendChunk(session->outputFifo,chunk,0);
#ifdef DEBUGSESSIONS
#ifdef USELOGIN
            syslog(LOG_DAEMON|LOG_NOTICE,
                   "Read from pty \"%s\" failed (errno=%d)",
                   session->ptyName,errno);
#else
            syslog(LOG_DAEMON|LOG_NOTICE,
                   "Read failed (errno=%d)",errno);
#endif
#endif
            markConnectionForDeletion(pty);
            return; }
        fifoAppendChunk(session->outputFifo,chunk,count); }
    processPendingPoll(connection,session);
    return;
}


static void sessionPollOut(int pty,Connection *connection,Session *session)
{
    // Our client process is willing to receive data
    int  i;

    while ((i = fifoWriteV(pty,session->inputFifo)) > 0)
        fifoDequeueChunk(session->inputFifo,i);
    if (i < 0) {
        if (errno == EAGAIN)
            return;
        // Write to PTY failed
#ifdef DEBUGSESSIONS
#ifdef USELOGIN
        syslog(LOG_DAEMON|LOG_NOTICE,
               "Write to pty \"%s\" failed (errno=%d)",
               session->ptyName,errno);
#else
        syslog(LOG_DAEMON|LOG_NOTICE,
               "Write failed (errno=%d)",errno);
#endif
#endif
        markConnectionForDeletion(pty);
        return; }
    return;
}


static void sessionPollHup(int pty,Connection *connection,Session *session)
{
    // Something terrible has happened...
#ifdef DEBUGSESSIONS
#ifdef USELOGIN
    syslog(LOG_DAEMON|LOG_NOTICE,"Slave end of pty \"%s\" has been closed",
           session->ptyName);
#else
    syslog(LOG_DAEMON|LOG_NOTICE,"Slave end has been closed");
#endif
#endif
    markConnectionForDeletion(pty);
    return;
}


static void sessionDestroy(int pty,Connection *connection,Session *session)
{
#ifdef USELOGIN
    struct utmpx *utmpx;
#endif

    // Notify CGI that the connection has been disconnected
    if (session->pendingPoll != NULL) {
        fAbortCGI(session->pendingPoll,404,
                  "Connection closed!\n"
                  "Connection closed by foreign host");
        destroyFifo(session->pendingPoll); /* decrease use count */
        session->pendingPoll = NULL; }
#ifdef USELOGIN
    if (!strncmp(session->ptyName,"/dev/tty",8)) {
        // If this is an old-style pty, reset permissions
        struct group *gr;
        int          ttygid;
        if ((gr = getgrnam("tty")) != NULL)
            ttygid = gr->gr_gid;
        else
            ttygid = -1;
        setreuid(-1,euid);
        chmod(session->ptyName,0666);
        chown(session->ptyName,0,ttygid);
        setreuid(-1,uid); }
    // Remove utmp and wtmp entry; the latter has been added by /bin/login
    utmpx          = &session->utmpx;
    utmpx->ut_type = DEAD_PROCESS;
    memset(utmpx->ut_user,0,sizeof(utmpx->ut_user));
    memset(utmpx->ut_host,0,sizeof(utmpx->ut_host));
    memset(&utmpx->ut_tv, 0,sizeof(utmpx->ut_tv));
    setreuid(-1,euid);
    setutxent();
    pututxline(utmpx);
    time(&utmpx->ut_tv.tv_sec);
    updwtmpx(WTMP_FILENAME,utmpx);
    endutxent();
    setreuid(-1,uid);
#endif
    // Release pty
    close(pty);
    // Remove old fifos
    destroyFifo(session->inputFifo);
    destroyFifo(session->outputFifo);
    return;
}


static void sessionDoTimeOut(int pty,Connection *connection,Session *session)
{
    if (session->pendingPoll != NULL) {
        // Queued poll event timed out; return zero bytes
        // and wait for new poll request
        static const char *hdr = "Content-type: "WLCONTENTTYPE"\r\n"
                                 "\r\n"DATA;
        char              *chunk = fifoCreateChunk(strlen(hdr));
        memcpy(chunk,hdr,strlen(hdr));
        fifoAppendChunk(session->pendingPoll,chunk,strlen(hdr));
        fifoFlushAndClose(session->pendingPoll);
        destroyFifo(session->pendingPoll); /* decrease use count */
        session->pendingPoll = NULL;
        setConnectionTimeOut(connection,CONNTIMEOUT); }
    else {
        // Connection keep-alive timed out; close session
#ifdef DEBUGSESSIONS
#ifdef USELOGIN
        syslog(LOG_DAEMON|LOG_NOTICE,
               "Connection timed out for pty \"%s\"",
               session->ptyName);
#else
        syslog(LOG_DAEMON|LOG_NOTICE,"Connection timed out");
#endif
#endif
        markConnectionForDeletion(pty); }
    return;
}


static void sessionReapProcess(int pty,Connection *connection,
                               Session *session,int pid)
{
    if (session->pid == pid) {
#ifdef DEBUGSESSIONS
#ifdef USELOGIN
        syslog(LOG_DAEMON|LOG_NOTICE,
               "Session leader (pid=%d) for pty \"%s\" died",
               pid,session->ptyName);
#else
        syslog(LOG_DAEMON|LOG_NOTICE,"Session leader (pid=%d) died",pid);
#endif
#endif
        markConnectionForDeletion(pty); }
    return;
}


static void sessionServeRequest(int sessionId,Connection *connection,
                                Session *session,FIFO *inputFifo,
                                int sequenceNumber,FIFO *outputFifo)
{
    // CGI is being POLLed for more output from our child process
    if (!fifoStrCmp(outputFifo,POLL)) {
#ifdef DEBUGREQUESTS
        syslog(LOG_DEBUG|LOG_DAEMON,"POLL: %d, %d",sessionId,sequenceNumber);
#endif
        // Verify that this is the next packet for this session
        session->outputSequenceNumber =
            (session->outputSequenceNumber + 1) & 0xFFFFFFFFUL;
        if (session->outputSequenceNumber != sequenceNumber) {
            fAbortCGI(inputFifo,400,"Illegal Request\r\n"
                                    "Sequence number mismatch!");
            markConnectionForDeletion(sessionId);
            return; }
        if (session->pendingPoll != NULL) {
            static const char *hdr = "Content-type: "WLCONTENTTYPE"\r\n"
                                     "\r\n"DATA;
            char              *chunk = fifoCreateChunk(strlen(hdr));
            memcpy(chunk,hdr,strlen(hdr));
            fifoAppendChunk(session->pendingPoll,chunk,strlen(hdr));
            fifoFlushAndClose(session->pendingPoll);
            destroyFifo(session->pendingPoll); /* decrease use count */ }
        session->pendingPoll = inputFifo;
        dupFifo(session->pendingPoll); /* increase use count */
        setConnectionTimeOut(connection,POLLTIMEOUT);
        processPendingPoll(connection,session);
        return; }
    // CGI is provided input DATA to be passed to our child process
    else if (!fifoStrCmp(outputFifo,DATA)) {
        static const char *hdr = "Content-type: "WLCONTENTTYPE"\r\n"
                                 "\r\n"DATA;
        char              *chunk;
        size_t            count;
#ifdef DEBUGREQUESTS
        syslog(LOG_DEBUG|LOG_DAEMON,"DATA: %d, %d",sessionId,sequenceNumber);
#endif
        // Verify that this is the next packet for this session
        if (session->inputSequenceNumber != sequenceNumber) {
            fAbortCGI(inputFifo,400,"Illegal Request\r\n"
                                    "Sequence number mismatch!");
            markConnectionForDeletion(sessionId);
            return; }
        session->inputSequenceNumber =
            (session->inputSequenceNumber + 1) & 0xFFFFFFFFUL;
        // Acknowledge request
        chunk = fifoCreateChunk(strlen(hdr));
        memcpy(chunk,hdr,strlen(hdr));
        fifoAppendChunk(inputFifo,chunk,strlen(hdr));
        fifoFlushAndClose(inputFifo);
        // Queue all data and wait for PTY to be ready to receive it
        fifoDequeueChunk(outputFifo,strlen(DATA));
        while ((chunk = fifoGetFirstChunk(outputFifo,&count)) != NULL) {
            char *newChunk = fifoCreateChunk(count);
            memcpy(newChunk,chunk,count);
            fifoAppendChunk(session->inputFifo,newChunk,count);
            fifoDequeueChunk(outputFifo,count); }
        return; }
    // CGI is asked to DISCONNECT and terminate all processes in this session
    else if (!fifoStrCmp(outputFifo,DISCONNECT)) {
#ifdef DEBUGREQUESTS
        syslog(LOG_DEBUG|LOG_DAEMON,"DISCONNECT: %d",sessionId);
#endif
        markConnectionForDeletion(sessionId);
        fAbortCGI(inputFifo,400,"Connection closed!\r\n"
                                "Connection closed by user");
        return; }
    else {
#ifdef DEBUGREQUESTS
        syslog(LOG_DEBUG|LOG_DAEMON,"UNKNOWN: %d, %d",
               sessionId,sequenceNumber);
#endif
        // Received an unknown request (protocol violation)
        markConnectionForDeletion(sessionId);
        fAbortCGI(inputFifo,400,"Illegal Request\r\n"
                                "Unknown data type received!"); }
    return;
}


Fnc sessionVTable[VTABSIZE] = {
    (Fnc)sessionPollIn,(Fnc)sessionPollOut,(Fnc)sessionPollHup,
    (Fnc)sessionDestroy,(Fnc)sessionDoTimeOut,(Fnc)sessionReapProcess,
    (Fnc)sessionServeRequest };

/*
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
