#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <syslog.h>
#include <unistd.h>

#include "connectionmgr.h"
#include "fifo.h"
#include "server.h"
#include "session.h"
#include "util.h"
#include "shellinaboxd.h"


int        uid,euid;
const char *daemonPath;


void fAbortCGI(FIFO *fifo,int errCode,char *msg,...)
{
    char    *buf;
    va_list argPtr;
    size_t  len;
    char    *chunk;
    
    va_start(argPtr,msg);
    buf = vAbortCGI(errCode,msg,argPtr);
    va_end(argPtr);
    len   = strlen(buf);
    chunk = fifoCreateChunk(len);
    memmove(chunk,buf,len);
    fifoAppendChunk(fifo,chunk,len);
    free(buf);
    fifoFlushAndClose(fifo);
    return;
}


int noFnc(void)
{
    return(0);
}


static void destroyServerData(int pty,Connection *connection,
                              ServerData *serverData)
{
    if (serverData != NULL) {
        FNC(serverData,Destroy)(pty,connection,&serverData->u);
        free(serverData); }
    return;
}


static void serverFnc(void)
{
    struct sockaddr_un sockAddr;
    int                serverSocket;
    Connection         *connection;

    // Open a new socket and listen for requests from the CGI
    if (strlen(daemonPath) + strlen(UNIXSOCKET) + 1 > UNIX_PATH_MAX) {
        errno = 0;
    sockErr:
        syslog(LOG_DAEMON|LOG_WARNING,"Internal Error; "
              "cannot open socket \"%s%s\" (errno=%d)!",daemonPath,UNIXSOCKET);
        _exit(1); }
    sockAddr.sun_family = AF_UNIX;
    strcat(strcpy(sockAddr.sun_path,daemonPath),UNIXSOCKET);
    unlink(sockAddr.sun_path);
    if ((serverSocket = socket(PF_UNIX,SOCK_STREAM,0)) < 0 ||
        bind(serverSocket,(struct sockaddr *)&sockAddr,sizeof(sockAddr)) ||
        listen(serverSocket,SOMAXCONN) ||
        fcntl(serverSocket,F_SETFL,(long)(O_NONBLOCK|O_RDWR)) != 0) {
        goto sockErr; }
    // Signal parent process that we are done initializing
    close(0);
    // Prepare to enter main loop
    initConnections();
    serverCreate(serverSocket);
    do {
        int i,pollRc,pid;
        // Wait for new file event and/or timeout
        pollRc = pollConnections();
        // Reap old processes
        while ((pid = waitpid(-1,&i,WNOHANG)) > 0) {
            if (WIFEXITED(i) || WIFSIGNALED(i)) {
                ServerData *serverData;
                int        pty;
                for (connection = firstConnection(&pty,&serverData);
                     connection != NULL;
                     connection = nextConnection(connection,&pty,&serverData))
                    FNC(serverData,ReapProcess)(pty,connection,
                                                &serverData->u,pid); } }
        if (pollRc == 0) {
            ServerData *serverData;
            int        pty;
            // Find timed-out pending polls and connections
            for (connection = firstConnection(&pty,&serverData);
                 connection != NULL;
                 connection = nextConnection(connection,&pty,&serverData)) {
                if (isTimedOut(connection))
                    FNC(serverData,DoTimeOut)(pty,connection,&serverData->u);}}
        else {
            ServerData *serverData;
            int        pty;
            for (connection = firstConnection(&pty,&serverData);
                 pollRc > 0 && connection != NULL;
                 connection = nextConnection(connection,&pty,&serverData)) {
                int revents = getConnectionEvents(connection);
                if (revents == 0)
                    continue;
                pollRc--;
                // Find all  pty related events
                if (revents & POLLIN)
                    FNC(serverData,PollIn)(pty,connection,&serverData->u);
                if (revents & POLLOUT)
                    FNC(serverData,PollOut)(pty,connection,&serverData->u);
                if (revents & (POLLERR|POLLHUP|POLLNVAL))
                    FNC(serverData,PollHup)(pty,connection,&serverData->u); } }
        // Remove marked connections, rebuild data structures, compute timeouts
    } while (deleteMarkedConnections(destroyServerData) > 1);
    return;
}


static int setLimit(int resource,int limit)
{
    struct rlimit oldRlim;
    struct rlimit newRlim;

    if (getrlimit(resource,&oldRlim) != 0)
        return(-1);
    if (oldRlim.rlim_cur == limit)
        return(0);
    newRlim.rlim_cur = limit;
    newRlim.rlim_max = oldRlim.rlim_max;
    if (setrlimit(resource,&newRlim) == 0)
        return(0);
    if (limit == RLIM_INFINITY || limit > oldRlim.rlim_cur) {
        newRlim.rlim_cur = oldRlim.rlim_max;
        newRlim.rlim_max = oldRlim.rlim_max;
        if (setrlimit(resource,&newRlim) == 0)
            return(0); }
    newRlim.rlim_cur = oldRlim.rlim_cur;
    newRlim.rlim_max = oldRlim.rlim_max;
    return(setrlimit(resource,&newRlim));
}


int main(int argc,char *argv[])
{
    char *pPtr;
    int  gid;

    // Simple sanity check to prevent shellinaboxd from being accidentally
    // mistaken with a CGI program
    if (argc != 2 || strcmp(argv[1],DAEMONPATH)) {
        syslog(LOG_DAEMON|LOG_WARNING,
               "Shellinaboxd must be called from shellinabox.cgi");
        exit(1);
        return(1); }
    // Reset limits that the web browser might have imposed on us; if the
    // system uses limits, then /bin/login will set them again
    setLimit(RLIMIT_CORE,0);
    setLimit(RLIMIT_CPU,RLIM_INFINITY);
    setLimit(RLIMIT_DATA,RLIM_INFINITY);
    setLimit(RLIMIT_FSIZE,RLIM_INFINITY);
    // Some reasonable default values
    setLimit(RLIMIT_NOFILE,1024);
    setLimit(RLIMIT_NPROC,256);
    setLimit(RLIMIT_STACK,8*1024*1024);
    // Leave MEMLOCK and RSS sizes alone. They affect system performance and
    // are hopefully set correctly.
    //   setLimit(RLIMIT_MEMLOCK,RLIM_INFINITY);
    //   setLimit(RLIMIT_RSS,RLIM_INFINITY);
    // Drop privileges. Warning! We keep saved-uid=0, so that we can reassert
    // permissions when we need them
    uid  = getuid();
    euid = geteuid();
    gid  = getgid();
#ifdef USELOGIN
    if (euid != 0) {
        syslog(LOG_DAEMON|LOG_WARNING,"Shellinaboxd needs to be setuid root");
        exit(1); }
#endif
    if (uid == 0
#ifndef USELOGIN
        || euid == 0
#endif
        ) {
        syslog(LOG_DAEMON|LOG_WARNING,
               "Shellinaboxd should not be executed by root"
#ifdef USELOGIN
               ", but it must be setuid root"
#endif
               );
        exit(1); }
#ifdef USELOGIN
    setreuid(-1,uid);
#else
    setreuid(uid,uid);
#endif
    setregid(gid,gid);
    // Compute the physical path where this program lives
    // following symbolic links
    daemonPath = resolveSymbolicLinks(argv[0]);
    if (daemonPath == NULL) {
        syslog(LOG_DAEMON|LOG_WARNING,"Too many symbolic links; "
                                      "cannot locate the shellinaboxd binary");
        exit(1); }
    pPtr = strrchr(daemonPath,'/');
    if (pPtr != NULL)
        *pPtr = '\000';
    // Initialize random number generator for sequence numbers
    srandom(time(NULL));
    // Start the server
    serverFnc();
    return(0);
}

/*
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
