#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#undef  DEBUGSHAPES


enum { UNKNOWN, RELEVANT, IRRELEVANT };
enum { FREE, PRIMARY, SUBSTITUTE, FONTPOS };


typedef struct Segment {
    int  relevant;
    int  rectangle;
    int  x1,y1,x2,y2,a,ra; // Rectangle
    int  x,y,dx,dy,l,rl;   // Line segment
} Segment;

// The segment data structure contains (in this order):
//  - rectangles that haven't been examined yet (UNKNOWN)
//  - line segments that haven't been examined yet (UNKNOWN)
//  - rectangles that are part of the result set (RELEVANT)
//  - line segments that are part of the result set (RELEVANT)
//  - rectangles that have been dropped from the result set (IRRELEVANT)
//  - line segments that have been dropped from the result set (IRRELEVANT)
// Within each group, entries are sorted according to how likely they
// are to contribute to an optimal result set.
static Segment        *segments   = NULL;
static int            numSegments = 0;
static unsigned short **allGlyphs;
static int            allGlyphsStatus[65536];
static int            numGlyphs   = 0;


// By selecting a sorting scheme that heuristically prefers
// segments/rectangles that are more likely to result in efficient
// drawing operations, we can direct the algorithm towards a
// more optimal result set.
static int cmpSegments(const void *arg1,const void *arg2)
{
    Segment *s1 = (Segment *)arg1;
    Segment *s2 = (Segment *)arg2;
    int     c1,c2;

    // Sort relevant segments to the end
    c1 = s1->relevant;
    c2 = s2->relevant;
    if (c1 > c2)          return(1);
    else if (c1 < c2)     return(-1);
    // Prefer rectangles
    c1 = s1->rectangle;
    c2 = s2->rectangle;
    if (c1 && !c2)        return(-1);
    else if (!c1 && c2)   return(1);
    if (c1) {
        // Sort by relevant area of rectangle
        c1 = s1->ra;
        c2 = s2->ra;
        if (c1 < c2)      return(1);
        else if (c1 > c2) return(-1);
        // Sort by height
        c1 = s1->y2 - s1->y1;
        c2 = s2->y2 - s2->y1;
        if (c1 < c2)      return(1);
        else if (c1 > c2) return(-1);
        // Sort by topmost y coordinate
        c1 = s1->y1;
        c2 = s2->y1;
        if (c1 > c2)      return(1);
        else if (c1 < c2) return(-1);
        // Sort by leftmost x coordinate
        c1 = s1->x1;
        c2 = s2->x1;
        if (c1 > c2)      return(1);
        else if (c1 < c2) return(-1);
        return(0);
    } else {
        // Sort by relevant length of segment
        c1 = s1->rl;
        c2 = s2->rl;
        if (c1 < c2)      return(1);
        else if (c1 > c2) return(-1);
        // Sort by angle
        c1 = (s1->dx == 0 || s1->dy == 0);
        c2 = (s2->dx == 0 || s2->dy == 0);
        if (c1 < c2)      return(1);
        else if (c1 > c2) return(-1);
        c1 = (s1->dx == 0);
        c2 = (s2->dx == 0);
        if (c1 < c2)      return(1);
        else if (c1 > c2) return(-1);
        c1 = (s1->dy == 0);
        c2 = (s2->dy == 0);
        if (c1 < c2)      return(1);
        else if (c1 > c2) return(-1);
        c1 = (s1->dx)*100+(s1->dy);
        c2 = (s2->dx)*100+(s2->dy);
        if (c1 < c2)      return(1);
        else if (c1 > c2) return(-1);
        // Sort by topmost y coordinate
        c1 = s1->dy < 0 ? s1->y + s1->l*s1->dy : s1->y;
        c2 = s2->dy < 0 ? s2->y + s2->l*s2->dy : s2->y;
        if (c1 > c2)      return(1);
        else if (c1 < c2) return(-1);
        // Sort by leftmost x coordinate
        c1 = s1->dx < 0 ? s1->x + s1->l*s1->dx : s1->x;
        c2 = s2->dx < 0 ? s2->x + s2->l*s2->dx : s2->x;
        if (c1 > c2)      return(1);
        else if (c1 < c2) return(-1);
        return(0); }
}


static void addSegment(int x,int y,int dx,int dy,int l)
{
    Segment newNode;

    memset(&newNode,-1,sizeof(newNode));
    newNode.relevant = UNKNOWN;
    newNode.rectangle= 0;
    newNode.x        = x;
    newNode.y        = y;
    newNode.dx       = dx;
    newNode.dy       = dy;
    newNode.l        = l;
    newNode.rl       = l;
    if (numSegments > 0 &&
        bsearch(&newNode,segments,numSegments,sizeof(Segment),cmpSegments))
        return;
    segments = realloc(segments,sizeof(Segment)*++numSegments);
    memmove(segments+numSegments-1,&newNode,sizeof(Segment));
    qsort(segments,numSegments,sizeof(Segment),cmpSegments);
    return;
}


static void addRectangle(int x1,int y1,int x2,int y2)
{
    Segment newNode;

    memset(&newNode,-1,sizeof(newNode));
    newNode.relevant = UNKNOWN;
    newNode.rectangle= 1;
    newNode.x1       = x1;
    newNode.y1       = y1;
    newNode.x2       = x2;
    newNode.y2       = y2;
    newNode.a        = (x2-x1+1) * (y2-y1+1);
    newNode.ra       = newNode.a;
    if (numSegments > 0 &&
        bsearch(&newNode,segments,numSegments,sizeof(Segment),cmpSegments))
        return;
    segments = realloc(segments,sizeof(Segment)*++numSegments);
    memmove(segments+numSegments-1,&newNode,sizeof(Segment));
    qsort(segments,numSegments,sizeof(Segment),cmpSegments);
    return;
}


static void removeAllSegments(void)
{
    if (numSegments > 0) {
        free(segments);
        segments    = NULL;
        numSegments = 0; }
    return;
}

// Find and enter a line segments starting from a given point in pointing
// towards a given direction
static void findSegment(int **img,int x,int y,int dx,int dy,int w,int h)
{
    int x2,y2;
    int l = 1;

    if (x-dx >= 0 && x-dx < w && 
        y-dy >= 0 && y-dy < h &&
        img[x-dx][y-dy])
        return;
    for (x2 = x, y2 = y;
         x2+dx >= 0 && x2+dx < w &&
         y2+dy >= 0 && y2+dy < h &&
         img[x2+dx][y2+dy];
         x2 += dx, y2 += dy, l++);
    if (l == 1)
        dx =
        dy = 0;
    addSegment(x,y,dx,dy,l);
    return;
}


// Recursively, find all rectangles containing a given point
static int findRectangle(int **img,int x1,int y1,int x2,int y2,int w,int h,
                         int *f)
{
    int ok = 0;
    int x,y,b = 0;

    if (x1 > 0) {
        for (y = y1; y <= y2; y++)
            if (!img[x1-1][y])
                goto r1;
        b  |= 1; }
 r1:if (y1 > 0) {
        for (x = x1; x <= x2; x++)
            if (!img[x][y1-1])
                goto r2;
        b |= 2; }
 r2:if (x2 < w-1) {
        for (y = y1; y <= y2; y++)
            if (!img[x2+1][y])
                goto r3;
        b |= 4; }
 r3:if (y2 < h-1) {
        for (x = x1; x <= x2; x++)
            if (!img[x][y2+1])
                goto r4;
        b |= 8; }
 r4:
    *f = b;
    if (b & 1) {
        ok |= findRectangle(img,x1-1,y1  ,x2  ,y2  ,w,h,&b);
        b = *f & ~b; }
    if (b & 2) {
        ok |= findRectangle(img,x1  ,y1-1,x2  ,y2  ,w,h,&b);
        b = *f & ~b; }
    if (b & 4) {
        ok |= findRectangle(img,x1  ,y1  ,x2+1,y2  ,w,h,&b);
        b = *f & ~b; }
    if (b & 8)
        ok |= findRectangle(img,x1  ,y1  ,x2  ,y2+1,w,h,&b);
    if (x1 == x2 || y1 == y2)
        return(0);
    if (!ok)
        addRectangle(x1,y1,x2,y2);
    return(1);
}


// Convert a degenerate rectangle to a line segment
static void convertToSegment(int **img,Segment *s)
{
    int l,x,y;

    if (!s->rectangle ||
        (s->x1 != s->x2 && s->y1 != s->y2))
        return;
    s->rectangle = 0;
    s->x         = s->x1;
    s->y         = s->y1;
    s->dx        = s->x2 > s->x1 ? 1 : 0;
    s->dy        = s->y2 > s->y1 ? 1 : 0;
    s->l         = s->dx ? s->x2 - s->x1 + 1 : s->y2 - s->y1 + 1;
    s->rl        = 0;
    for (l = s->l, x = s->x, y = s->y; l-- > 0; x += s->dx, y += s->dy)
        if (img[x][y] == 1)
            s->rl++;
    if (s->rl <= 0)
        s->relevant = IRRELEVANT;
    return;
}


// Mark all UNKNOWNn entries as either RELEVANT or IRRELEVANT; return the
// list of RELEVANT drawing operations
static Segment *markRelevantSegments(int **img,int *num,int w,int h)
{
    int x,y,l,n;
    int isRelevant = 0;

    if (numSegments == 0) {
        *num = 0;
        return(NULL); }
    // Process all segments/rectangles until they are classified as
    // RELEVANT or IRRELEVANT
    while (segments->relevant == UNKNOWN) {
        // Check is the segment with the highest priority still contributes
        // any new data
        if (segments->rectangle) {
            for (x = segments->x1; x <= segments->x2; x++)
                for (y = segments->y1; y <= segments->y2; y++)
                    if (img[x][y]++ == 1)
                        isRelevant++; }
        else
            for (l = segments->l, x = segments->x, y = segments->y;
                 l-- > 0; x += segments->dx, y += segments->dy)
                if (img[x][y]++ == 1)
                    isRelevant++;
        if (isRelevant) {
            Segment *s;
            int      i;
            // Mark this segment as relevant and recompute data for all
            // other unknown segments
            segments->relevant = RELEVANT;
            for (i = 1;i < numSegments && (s = segments+i)->relevant == 0;i++)
                if (s->rectangle) {
                    int r, c;
                    // Prune rectangle on all sides
                    do {
                        c = 0;
                        for (r = 0, y = s->y1; y <= s->y2; y++)
                            if (img[s->x1][y] <= 1) { r++; break; }
                        if (!r) { if (s->x1++ >= s->x2) break; c++; }
                        for (r = 0, x = s->x1; x <= s->x2; x++)
                            if (img[x][s->y1] <= 1) { r++; break; }
                        if (!r) { if (s->y1++ >= s->y2) break; c++; }
                        for (r = 0, y = s->y1; y <= s->y2; y++)
                            if (img[s->x2][y] <= 1) { r++; break; }
                        if (!r) { if (s->x2-- <= s->x1) break; c++; }
                        for (r = 0, x = s->x1; x <= s->x2; x++)
                            if (img[x][s->y2] <= 1) { r++; break; }
                        if (!r) { if (s->y2-- <= s->y1) break; c++; }
                    } while (c);
                    // Remove rectangle if nothing relevant is left
                    if (s->x1 > s->x2 || s->y1 > s->y2)
                        s->relevant = IRRELEVANT;
                    // Convert to line segment if rectangle is degenerate
                    else if (s->x1 == s->x2 || s->y1 == s->y2)
                        convertToSegment(img,s);
                    else {
                        // Count relevant area
                        isRelevant = 0;
                        for (x = s->x1; x <= s->x2; x++)
                            for (y = s->y1; y <= s->y2; y++)
                                if (img[x][y] <= 1)
                                    isRelevant++;
                        s->ra = isRelevant;
                        if (!isRelevant)
                            s->relevant = IRRELEVANT; }
                } else {
                    // Prune line segments from both ends
                    for (l = s->l, x = s->x, y = s->y; l-- > 0;
                         x += s->dx, y += s->dy)
                        if (img[x][y] == 1)
                            break;
                        else {
                            s->x += s->dx;
                            s->y += s->dy;
                            s->l--;
                            s->rl--; }
                    isRelevant = 0;
                    s->rl      = s->l;
                    for (l= s->l,x = s->x + s->dx*(l-1),y = s->y + s->dy*(l-1);
                         l-- > 0; x -= s->dx, y -= s->dy)
                        if (img[x][y] == 1)
                            isRelevant++;
                        else {
                            if (!isRelevant)
                                s->l--;
                            s->rl--; }
                    // Mark line segments as irrelevant if they have shrunk
                    // to zero
                    if (s->l <= 0)
                        s->relevant = IRRELEVANT; } }
        else {
            // Remove segment because it doesn't contribute new data
            segments->relevant = IRRELEVANT;
            // Clean up image
            if (segments->rectangle)
                for (x = segments->x1; x <= segments->x2; x++)
                    for (y = segments->y1; y <= segments->y2; y++)
                        img[x][y]--;
            else
                for (l = segments->l, x = segments->x, y = segments->y;
                     l-- > 0; x += segments->dx, y += segments->dy)
                    img[x][y]--; }
        // Reorganize list of segments/rectangles
        qsort(segments,numSegments,sizeof(Segment),cmpSegments); }
    // We now have to postprocess the list of RELEVANT rectangles because
    // they might require some tuning after the RELEVANT line segments have
    // been added:
    // Normalize image
    for (x = 0; x < w; x++)
        for (y = 0; y < h; y++)
            img[x][y] = !!img[x][y];
    // Enter all line segments into the image
    for (n = 0; n < numSegments && segments[n].relevant < IRRELEVANT &&
                                   segments[n].rectangle; n++);
    for (; n < numSegments && segments[n].relevant < IRRELEVANT; n++)
        for (l = segments[n].l, x = segments[n].x, y = segments[n].y; l-- > 0;
             x += segments[n].dx, y += segments[n].dy)
            img[x][y]++;
    // Check if rectangles need additional pruning
    for (n = 0; n < numSegments && segments[n].relevant < IRRELEVANT &&
                                   segments[n].rectangle; n++) {
        int r, c;
        // Prune rectangle on all sides
        do {
            c = 0;
            for (r = 0, y = segments[n].y1; y <= segments[n].y2; y++)
                if (img[segments[n].x1][y] <= 1) { r++; break; }
            if (!r) { if (segments[n].x1++ >= segments[n].x2) break; c++; }
            for (r = 0, x = segments[n].x1; x <= segments[n].x2; x++)
                if (img[x][segments[n].y1] <= 1) { r++; break; }
            if (!r) { if (segments[n].y1++ >= segments[n].y2) break; c++; }
            for (r = 0, y = segments[n].y1; y <= segments[n].y2; y++)
                if (img[segments[n].x2][y] <= 1) { r++; break; }
            if (!r) { if (segments[n].x2-- <= segments[n].x1) break; c++; }
            for (r = 0, x = segments[n].x1; x <= segments[n].x2; x++)
                if (img[x][segments[n].y2] <= 1) { r++; break; }
            if (!r) { if (segments[n].y2-- <= segments[n].y1) break; c++; }
        } while (c);
        // Remove rectangle if nothing relevant is left
        if (segments[n].x1 > segments[n].x2 || segments[n].y1 > segments[n].y2)
            segments[n].relevant = IRRELEVANT;
        // Convert to line segment if rectangle is degenerate
        else if (segments[n].x1 == segments[n].x2 ||
                 segments[n].y1 == segments[n].y2)
            convertToSegment(img,segments + n);
    }
    // Reorganize list of segments
    qsort(segments,numSegments,sizeof(Segment),cmpSegments);
    // Normalize image so that the data structure is left unchanged from
    // when the function was called
    for (x = 0; x < w; x++)
        for (y = 0; y < h; y++)
            img[x][y] = !!img[x][y];
    // Count number of relevant segments/rectangles and return result set
    for (n = 0; n < numSegments && segments[n].relevant < IRRELEVANT; n++);
    *num = n;
    return(n > 0 ? segments : NULL);
}


static void vectorize(int uniCode,int isPrimaryGlyph,int **img,int glyphWidth,
                      int glyphHeight)
{
    int            x,y,num,i;
    Segment        *shape;
    unsigned short *ins;

    // Check whether this is a new glyph for the given unicode position; there
    // are three priorities:
    //  - PRIMARY: glyph is the official shape for this unicode character
    //  - SUBSTITUTE: glyph can be used if there is no other shape available
    //    for this unicode character (e.g. it lacks required accents but is
    //    otherwise intelligible
    //  - FONTPOS: glyph happens to be at this character position in the first
    //    font file provided on the program's command line. This feature is
    //    intended to initialize unused character positions in the range 0..255
    //    to the defaults that they have in the ISO Latin 1 font.
    if (allGlyphsStatus[uniCode] == PRIMARY ||
        (allGlyphsStatus[uniCode] == SUBSTITUTE && isPrimaryGlyph <= 0) ||
        (allGlyphsStatus[uniCode] == FONTPOS    && isPrimaryGlyph <  0))
        return;
    if (allGlyphsStatus[uniCode] == FREE)
        numGlyphs++;
    // Examine every set pixel in the shape and find all line segments and
    // rectangles starting from there
    for (y = 0; y < glyphHeight; y++) {
        for (x = 0; x < glyphWidth; x++) {
            if (img[x][y]) {
                int f;
                findRectangle(img,x,y,x,y,glyphWidth,glyphHeight,&f);
                findSegment(img,x,y,0, 1,glyphWidth,glyphHeight);
                findSegment(img,x,y,1, 0,glyphWidth,glyphHeight);
                findSegment(img,x,y,1, 1,glyphWidth,glyphHeight);
                findSegment(img,x,y,1,-1,glyphWidth,glyphHeight); } } }
    shape = markRelevantSegments(img,&num,glyphWidth,glyphHeight);
#ifdef DEBUGSHAPES
    if (isPrimaryGlyph > 0) {
        fprintf(stderr,"\033c");
        {int color = 0;
        for (i = 0; i < num; i++) {
            int l;
            if (shape[i].rectangle)
                for (x = shape[i].x1; x <= shape[i].x2; x++)
                    for (y = shape[i].y1; y <= shape[i].y2; y++)
                        fprintf(stderr,"\033[%d;%dH\033[%d;%dm..\033[m",
                                y+2,2*x+10,(color%7)+41,(color/7)+30);
            else
                for (l = shape[i].l, x = shape[i].x, y = shape[i].y;
                     l-- > 0; x += shape[i].dx, y += shape[i].dy)
                    fprintf(stderr,"\033[%d;%dH\033[%d;%dm##\033[m",
                            y+2,2*x+10,(color%7)+41,(color/7)+30);
            color++; } }
        fprintf(stderr,"\033[21HU+%04X: (%d) ",uniCode,num);
        for (i = 0; i < num; i++) {
            int x1,y1,x2,y2;
            if (shape[i].rectangle) {
                x1 =  shape[i].x1                            & 0x7;
                y1 =  shape[i].y1                            & 0xF;
                x2 =  shape[i].x2                            & 0x7;
                y2 =  shape[i].y2                            & 0xF;
            } else {
                x1 =  shape[i].x                             & 0x7;
                y1 =  shape[i].y                             & 0xF;
                x2 = (shape[i].x+(shape[i].l-1)*shape[i].dx) & 0x7;
                y2 = (shape[i].y+(shape[i].l-1)*shape[i].dy) & 0xF; }
            fprintf(stderr,"%04X%s",(x1<<12)|(y1<<8)|(x2<<4)|(y2)|
                    (shape[i].rectangle?0x80:0x00),
                    i>=num-1?"":" ");
        }
        getchar(); }
#endif
    if (allGlyphsStatus[uniCode] != FREE)
        free(allGlyphs[uniCode]);
    // Convert list of RELEVANT drawing instructions to encoded coordinates
    // and add them to the list of known glyphs. Coordinates are stored in
    // the four nibbles of a sixteen bit word, supporting glyph sizes up
    // to 8x16. Bit 15 indicates the end of a sequence of coordinates; bit 7
    // indicates that the coordinates encode a rectangle (opposing corners)
    // if the bit is set, or that the coordinates encode a line segment
    // (start and end points) if the bit is cleared. Data is stored in
    // network order. The magic code 0x8080 is used for the empty glyph.
    allGlyphs[uniCode] = ins = calloc(sizeof(unsigned short),!num ? 1 : num);
    allGlyphs[uniCode][0] = 0x8080; // blank glyph
    for (i = 0; i < num; i++) {
        int x1,y1,x2,y2;
        if (shape[i].rectangle) {
            x1 =  shape[i].x1                            & 0x7;
            y1 =  shape[i].y1                            & 0xF;
            x2 =  shape[i].x2                            & 0x7;
            y2 =  shape[i].y2                            & 0xF;
        } else {
            x1 =  shape[i].x                             & 0x7;
            y1 =  shape[i].y                             & 0xF;
            x2 = (shape[i].x+(shape[i].l-1)*shape[i].dx) & 0x7;
            y2 = (shape[i].y+(shape[i].l-1)*shape[i].dy) & 0xF; }
        ins[i] = (x1<<12)|(y1<<8)|(x2<<4)|(y2)|
                 ((i == num-1) ? 0x8000 : 0x00) |
                 (shape[i].rectangle ? 0x80 : 0x00); }
    allGlyphsStatus[uniCode] = isPrimaryGlyph ? isPrimaryGlyph < 0
                               ? FONTPOS : PRIMARY : SUBSTITUTE;
    // Remove temporary storage
    removeAllSegments();
    segments = NULL;
    return;
}


int main(int argc,char *argv[]) {
    FILE  *fp;
    char  fontHeader[4];
    int   numberOfGlyphs;
    int   glyphHeight;
    char  *glyphs;
    int   n,m,o,i,j;
    int   cellWidth = -1,cellHeight = -1;

    allGlyphs = calloc(65536,sizeof(char *));
    for (n = 1; n < argc; n++) {
        // Read PSFU font file header
        if ((fp = fopen(argv[n],"r")) == NULL ||
            fread(fontHeader,sizeof(fontHeader),1,fp) != 1 ||
            fontHeader[0] != 0x36 ||
            fontHeader[1] != 0x04 ||
            (fontHeader[2] != 0x02 && fontHeader[2] != 0x03) ||
            fontHeader[3] <  6    || fontHeader[3] >  16) {
            fprintf(stderr,"Usage: %s fontfile.psfu ...\n",argv[0]);
            exit(1); }
        // Read glyphs
        numberOfGlyphs = (fontHeader[2]&1) ? 512 : 256;
        glyphHeight    = fontHeader[3];
        glyphs = malloc(numberOfGlyphs*glyphHeight);
        if (cellWidth < 0 || cellHeight < 0) {
            cellWidth  = 8;
            cellHeight = glyphHeight; }
        if (fread(glyphs,glyphHeight,numberOfGlyphs,fp) != numberOfGlyphs) {
            fprintf(stderr,"Cannot read glyphs!\n");
            exit(1); }
        // Loop over all glyphs
        for (i = j = 0; i < numberOfGlyphs; ) {
            char uniCodeBuffer[2];
            int  uniCode;
            // Read unicode entry
            if (fread(uniCodeBuffer,2,1,fp) != 1) {
                fprintf(stderr,"Cannot read unicode table!\n");
                exit(1); }
            uniCode = ((((int)uniCodeBuffer[0])     ) & 0x00FF) |
                      ((((int)uniCodeBuffer[1]) << 8) & 0xFF00);
            if (uniCode == 0xFFFF) {
                i++; j = 0; }
            else {
                int **img;
                int x,y;
                // Create internal representation of the glyphs shape. This
                // space is used to store intermediate results of calculations.
                img = malloc(8*(sizeof(int *)+glyphHeight*sizeof(int)));
                for (x = 0; x < 8; x++) {
                    img[x] = (int *)((char*)(img+8)+x*glyphHeight*sizeof(int));
                    for (y = 0; y < glyphHeight; y++) {
                        img[x][y] = !!(glyphs[i*glyphHeight+y]&(128 >> x)); } }
                // Compute primitive drawing operations
                vectorize(uniCode,!j++,img,8,glyphHeight);
                // For the first font file and the positions 0..255 initialize
                // unused unicode positions with the glyphs that correspond to
                // the ISO Latin 1 encoding
                if (i < 256)
                    vectorize(i,-1,img,8,glyphHeight);
                free(img); } }
        fclose(fp);
        free(glyphs); }
    if (numGlyphs <= 0) {
        fprintf(stderr,"No glyphs found\n");
        exit(1); }
    else if (cellWidth < 0 || cellHeight < 0) {
        fprintf(stderr,"Could not determine default glyph dimensions\n");
        exit(1); }
    // Write file header
    putchar(cellWidth);
    putchar(cellHeight);
    if (cellHeight <= 8)
        putchar(cellHeight-1);
    else if (cellHeight <= 14)
        putchar(cellHeight-2);
    else
        putchar(cellHeight-3);
    putchar((numGlyphs/256) & 0xFF);
    putchar( numGlyphs      & 0xFF);
    // Write all known glyphs
    for (i = n = m = o = 0; i < 65536; i++) {
        if (allGlyphsStatus[i] != FREE) {
            putchar((i >> 8) & 0xFF);
            putchar( i       & 0xFF);
            j = -1;
            do {
                n++;
                if (i >= 'a' && i <= 'z')      // 26*6
                    m += 6;
                else if (i >= 'A' && i <= 'Z') // 26*2
                    m += 2;
                else if (i > ' ' && i < 0x7f)  // 95*1
                    m++;
                j++;
                putchar((allGlyphs[i][j] >> 8) & 0xFF);
                putchar( allGlyphs[i][j]       & 0xFF);
            } while (!(allGlyphs[i][j] & 0x8000));
            if (j+1 > o)
                o = j+1;
        }
    }
    fprintf(stderr,"Wrote glyphs for %d distinct unicode positions\n"
                   "Default glyph dimension is %dx%d\n"
                   "Average number of instructions per glyph is %d\n"
                   "Weighed number of instructions per glyph is %d\n"
                   "Maximum number of instructions per glyph is %d\n",
            numGlyphs,cellWidth,cellHeight,
            (n+numGlyphs/2)/numGlyphs,(m+151)/303,o);
    return(0);
}

/*
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
